
__all__ = ["get_session_info"]


def get_session_info(identity_url, session_uid,
                     scope=None, permissions=None):
    """Call the identity_url to obtain information about the
       specified login session_uid. Optionally limit
       the scope and permissions for which these certs would
       be valid
    """
    from Acquire.Service import get_trusted_service as _get_trusted_service

    service = _get_trusted_service(identity_url)

    args = {"session_uid": session_uid}

    if scope is not None:
        args["scope"] = scope

    if permissions is not None:
        args["permissions"] = permissions

    response = service.call_function(function="get_session_info", args=args)

    try:
        del response["status"]
    except:
        pass

    try:
        del response["message"]
    except:
        pass

    from Acquire.Crypto import PublicKey as _PublicKey

    for key in response.keys():
        if key in ["public_key", "public_certificate"]:
            response[key] = _PublicKey.from_data(response[key])

    return response




__all__ = ["get_trusted_registry_service", "get_primary_registry_uid"]


def get_primary_registry_uid(service_uid):
    """Return the primary registry uid of the service with
       passed service_uid.
    """
    try:
        root = service_uid.split("-")[0]
        return "%s-%s" % (root, root)
    except:
        return "a0-a0"


def get_trusted_registry_service(registry_uid=None,
                                 service_uid=None, service_url=None):
    """Return the trusted service info for the registry with specified
       registry_uid, or get any trusted registry service using either
       'service_uid' or 'service_url' as a starting hint to
       locate a suitable registry
    """
    if service_uid is not None:
        # for the moment, just try to get one registry. Eventually we should
        # try to get several in case this one is down
        registry_uid = get_primary_registry_uid(service_uid)
        return get_trusted_registry_service(registry_uid=registry_uid)

    if service_url is not None:
        if service_url.find(".") != -1:
            # try the main acquire registry first
            return get_trusted_registry_service(registry_uid="a0-a0")
        else:
            # this must be testing...
            return get_trusted_registry_service(registry_uid="Z9-Z9")

    if registry_uid is None:
        raise PermissionError(
            "You must specify one of registry_uid, service_uid "
            "or service_url")

    from Acquire.Service import get_trusted_service as _get_trusted_service

    try:
        registry_service = _get_trusted_service(service_uid=registry_uid,
                                                autofetch=False)
    except:
        registry_service = None

    if registry_service is not None:
        if not registry_service.is_registry_service():
            from Acquire.Service import ServiceError
            raise ServiceError(
                "The requested service (%s) for %s is NOT a "
                "registry service!" % (registry_service, registry_uid))

        if registry_service.uid() != registry_uid:
            from Acquire.Service import ServiceError
            raise ServiceError(
                "Disagreement of UID (%s) is NOT the right registry service!" %
                registry_service)

        # everything is ok - we have seen this registry before
        return registry_service

    # boostrapping
    from Acquire.Registry import get_registry_details \
        as _get_registry_details

    details = _get_registry_details(registry_uid)

    from Acquire.Service import call_function as _call_function
    from Acquire.Service import Service as _Service
    from Acquire.Crypto import get_private_key as _get_private_key
    from Acquire.Crypto import PrivateKey as _PrivateKey
    from Acquire.Crypto import PublicKey as _PublicKey
    from Acquire.ObjectStore import bytes_to_string as _bytes_to_string
    from Acquire.ObjectStore import string_to_bytes as _string_to_bytes

    privkey = _get_private_key(key="registry")

    pubkey = _PublicKey.from_data(details["public_key"])
    pubcert = _PublicKey.from_data(details["public_certificate"])

    # ask the registry to return to us their latest details - use
    # a challenge-response to make sure that the response is
    # properly returned
    challenge = _PrivateKey.random_passphrase()
    encrypted_challenge = _bytes_to_string(pubkey.encrypt(challenge))
    args = {"challenge": encrypted_challenge,
            "fingerprint": pubkey.fingerprint()}

    result = _call_function(service_url=details["canonical_url"],
                            function=None,
                            args=args,
                            args_key=pubkey,
                            response_key=privkey,
                            public_cert=pubcert)

    if result["response"] != challenge:
        from Acquire.Service import ServiceError
        raise ServiceError(
            "The requested service (%s) failed to respond to the challenge!" %
            registry_service)

    registry_service = _Service.from_data(result["service_info"])

    if not registry_service.is_registry_service():
        from Acquire.Service import ServiceError
        raise ServiceError(
            "The requested service (%s) is NOT a registry service!" %
            registry_service)

    if registry_service.uid() != details["uid"]:
        from Acquire.Service import ServiceError
        raise ServiceError(
            "Disagreement of UID (%s) is NOT the right registry service!" %
            registry_service)

    # ok - we've got the registry - add this to the set of
    # trusted services so that we don't need to bootstrap from
    # the registry details again
    from Acquire.Service import trust_service as _trust_service
    _trust_service(registry_service)

    return registry_service





class PAR:
    """This class holds a pre-authenticated request to access
       a file, set of files or a drive. Anyone who holds a copy of this
       PAR can resolve it to recover a valid File or Drive object
       that will support access via the permissions of the PAR
    """
    def __init__(self, location=None, user=None, aclrule=None,
                 expires_datetime=None):
        """Construct a PAR for the specified location,
           authorised by the passed user, giving permissions
           according to the passed 'aclrule' (default is
           ACLRule.reader()).

           The passed 'expires_datetime' is the time at which
           this PAR will expire (by default within 24 hours)
        """
        self._location = None
        self._uid = None
        self._expires_datetime = None

        if location is None:
            return

        from Acquire.Client import Location as _Location
        if not isinstance(location, _Location):
            raise TypeError("The location must be type Location")

        if location.is_null():
            return

        from Acquire.Client import User as _User
        if not isinstance(user, _User):
            raise TypeError("The user must be type User")

        if not user.is_logged_in():
            raise PermissionError("The passed User must be logged in!")

        from Acquire.Client import ACLRule as _ACLRule

        if aclrule is None:
            aclrule = _ACLRule.reader()
        elif not isinstance(aclrule, _ACLRule):
            raise TypeError("The aclrule must be type ACLRule")

        if expires_datetime is None:
            from Acquire.ObjectStore import get_datetime_future \
                as _get_datetime_future
            expires_datetime = _get_datetime_future(days=1)
        else:
            from Acquire.ObjectStore import datetime_to_datetime \
                as _datetime_to_datetime
            expires_datetime = _datetime_to_datetime(expires_datetime)

        self._location = location
        self._expires_datetime = expires_datetime
        self._aclrule = aclrule

        from Acquire.Client import Authorisation as _Authorisation
        auth = _Authorisation(user=user,
                              resource="create_par %s" % self.fingerprint())

        from Acquire.Crypto import PrivateKey as _PrivateKey
        self._secret = _PrivateKey.random_passphrase()

        args = {"authorisation": auth.to_data(),
                "par": self.to_data(),
                "secret": self._secret}

        service = location.service()

        result = service.call_function(function="create_par",
                                       args=args)

        self._set_uid(result["par_uid"])

    def __str__(self):
        if self.is_null():
            return "PAR::null"
        elif not self.is_authorised():
            return "PAR::unauthorised"
        else:
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            return "PAR( %s, %s, expires %s )" % (
                    self._location.to_string(),
                    self._aclrule, _datetime_to_string(self._expires_datetime))

    def is_null(self):
        """Return whether or not this is null"""
        return self._location is None

    def is_authorised(self):
        """Return whether or not this has been authorised"""
        return self._uid is not None

    def assert_valid(self):
        """Assert that this is a valid and authorised PAR that
           has not yet expired
        """
        is_valid = True

        if self.is_null():
            is_valid = False
        elif not self.is_authorised():
            is_valid = False
        elif self.expired():
            is_valid = False

        if not is_valid:
            raise PermissionError("The PAR is not valid")

    def _set_uid(self, uid):
        """Internal function to set the UID of this PAR"""
        if self._uid is not None:
            raise PermissionError("You cannot set the UID twice!")

        self._uid = uid

    def uid(self):
        """Return the UID for this PAR"""
        if self.is_authorised():
            return self._uid
        else:
            return None

    def aclrule(self):
        """Return the ACL rule associated with this PAR"""
        if self.is_authorised():
            return self._aclrule
        else:
            from Acquire.Client import ACLRule as _ACLRule
            return _ACLRule.denied()

    def location(self):
        """Return the location for the Drive/File that is resolvable
           from this PAR
        """
        if self.is_authorised():
            return self._location
        else:
            return None

    def fingerprint(self):
        """Return a fingerprint that can be used to show that
           the user authorised the request to create this PAR
        """
        if self.is_null():
            return None
        else:
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            return "%s:%s:%s" % (self._location.fingerprint(),
                                 self._aclrule.fingerprint(),
                                 _datetime_to_string(self._expires_datetime))

    def service(self):
        """Return the service that authorised this PAR"""
        if self.is_null():
            return None
        else:
            return self._location.service()

    def service_url(self):
        """Return the URL of the service that authorised this PAR"""
        if self.is_null():
            return None
        else:
            return self._location.service_url()

    def service_uid(self):
        """Return the UID of the service that authorised this PAR"""
        if self.is_null():
            return None
        else:
            return self._location.service_uid()

    def secret(self):
        """Return the secret used to access this PAR, if this is held
           in this PAR
        """
        if self.is_null():
            return None
        else:
            return self._secret

    def resolve(self, secret=None):
        """Resolve this PAR into the authorised Drive or File object, ready
           for download, upload etc.

           If 'secret' is supplied, then this will use the supplied
           secret to unlock the PAR (sometimes they may be locked
           by a simple secret)
        """
        if not self.is_authorised():
            raise PermissionError(
                "You cannot resolve a PAR that has not been authorised!")

        service = self.service()

        if secret is None:
            try:
                secret = self._secret
            except:
                pass

        if secret is not None and len(secret) > 0:
            from Acquire.Crypto import Hash
            secret = Hash.multi_md5(self._uid, secret)
        else:
            secret = None

        result = service.call_function(function="resolve_par",
                                       args={"par_uid": self.uid(),
                                             "secret": secret})

        from Acquire.Client import StorageCreds as _StorageCreds
        creds = _StorageCreds(par=self, secret=secret)

        if result["type"] == "DriveMeta":
            from Acquire.Client import DriveMeta as _DriveMeta
            from Acquire.Client import Drive as _Drive
            return _Drive.open(_DriveMeta.from_data(result["data"]),
                               creds=creds)
        elif result["type"] == "FileMeta":
            from Acquire.Client import FileMeta as _FileMeta
            from Acquire.Client import File as _File
            return _File.open(_FileMeta.from_data(result["data"]),
                              creds=creds)
        elif result["type"] == "FileMetas":
            from Acquire.Client import FileMeta as _FileMeta
            from Acquire.Client import File as _File
            from Acquire.ObjectStore import string_to_list \
                as _string_to_list
            filemetas = _string_to_list(result["data"], _FileMeta)

            files = []
            for filemeta in filemetas:
                files.append(_File.open(filemeta, creds=creds))

            if len(files) == 1:
                return files[0]
            else:
                return files
        else:
            raise PermissionError("Returned wrong type? %s" % result["type"])

    def expires_when(self):
        """Return when this PAR expires (or expired)"""
        if not self.is_authorised():
            return None
        else:
            return self._expires_datetime

    def expired(self, buffer=30):
        """Return whether or not this PAR has expired"""
        return self.seconds_remaining(buffer=buffer) <= 0

    def seconds_remaining(self, buffer=30):
        """Return the number of seconds remaining before this PAR expires.
           This will return 0 if the PAR has already expired. To be safe,
           you should renew PARs if the number of seconds remaining is less
           than 60. This will subtract 'buffer' seconds from the actual
           validity to provide a buffer against race conditions (function
           says this is valid when it is not)

           Args:
                buffer (int, default=30): buffer PAR validity (seconds)
           Returns:
                datetime: Seconds remaining on PAR validity
        """
        if not self.is_authorised():
            return 0

        from Acquire.ObjectStore import get_datetime_now as _get_datetime_now

        buffer = float(buffer)

        if buffer < 0:
            buffer = 0

        now = _get_datetime_now()

        delta = (self._expires_datetime - now).total_seconds() - buffer

        if delta < 0:
            return 0
        else:
            return delta

    def to_data(self):
        """Return a json-serialisable dictionary of this PAR"""
        if self.is_null():
            return None

        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string

        data = {}

        data["location"] = self._location.to_data()
        data["aclrule"] = self._aclrule.to_data()
        data["expires_datetime"] = _datetime_to_string(self._expires_datetime)
        data["uid"] = self._uid

        return data

    @staticmethod
    def from_data(data):
        """Return a PAR constructed from the json-deserialised passed
           dictionary
        """
        if data is None or len(data) == 0:
            return PAR()

        f = PAR()

        from Acquire.Client import Location as _Location
        from Acquire.Client import ACLRule as _ACLRule
        from Acquire.ObjectStore import string_to_datetime \
            as _string_to_datetime

        f._location = _Location.from_data(data["location"])
        f._aclrule = _ACLRule.from_data(data["aclrule"])
        f._expires_datetime = _string_to_datetime(data["expires_datetime"])
        f._uid = data["uid"]

        return f




__all__ = ["StorageCreds"]


def _get_storage_url():
    """Function to discover and return the default storage url"""
    return "http://fn.acquire-aaai.com:8080/t/storage"


def _get_storage_service(storage_url=None):
    """Function to return the storage service for the system

       Args:
            storage_url (str, default=None): Storage URL to use
       Returns:
            Service: Service object
    """
    if storage_url is None:
        storage_url = _get_storage_url()

    from Acquire.Client import Service as _Service
    service = _Service(storage_url, service_type="storage")

    if not service.is_storage_service():
        from Acquire.Client import LoginError
        raise LoginError(
            "You can only use a valid storage service to get CloudDrive info! "
            "The service at '%s' is a '%s'" %
            (storage_url, service.service_type()))

    assert(service is not None)

    return service


class StorageCreds:
    """This class holds the credentials necessary to access
       Drives and Files. The credentials will either be a logged
       in user, or a valid PAR
    """
    def __init__(self, user=None, storage_service=None,
                 service_url=None, par=None, secret=None):
        """Create these credentials either from a logged-in
           user and associated storage service (or URL),
           or from a valid PAR with associated secret
        """
        self._user = None
        self._storage_service = None
        self._par = None
        self._secret = None

        if user is not None:
            from Acquire.Client import User as _User
            if not isinstance(user, _User):
                raise TypeError("The user must be type User")

            if not user.is_logged_in():
                raise PermissionError("The user must be logged in!")

            self._user = user

            if storage_service is None:
                storage_service = _get_storage_service(service_url)

            self._storage_service = storage_service
            assert(storage_service is not None)

        elif par is not None:
            from Acquire.Client import PAR as _PAR
            if not isinstance(par, _PAR):
                raise TypeError("par must be type PAR")

            if par.expired():
                raise PermissionError(
                    "The passed PAR is either invalid or expired!")

            self._par = par
            self._secret = secret
            self._storage_service = par.service()
            assert(self._storage_service is not None)

        if self._storage_service is not None:
            from Acquire.Storage import StorageService as _StorageService
            if not isinstance(self._storage_service, _StorageService):
                raise TypeError("storage service must be type StorageService")

    def is_user(self):
        """Return whether authentication is via a logged-in user"""
        return self._user is not None

    def is_par(self):
        """Return whether authentication is via a PAR"""
        return self._par is not None

    def user(self):
        if self._user is None:
            raise PermissionError(
                "There is no user associated with these credentials!")
        return self._user

    def par(self):
        if self._par is None:
            raise PermissionError(
                "There is no PAR associated with these credentials!")

        return self._par

    def secret(self):
        # null secrets are allowed
        return self._secret

    def storage_service(self):
        if self._storage_service is None:
            raise PermissionError(
                "There is no storage service associated with these "
                "credentials!")

        return self._storage_service




__all__ = ["OSParRegistry"]

_registry_key = "registry/pars"


class OSParRegistry:
    """This is a OSPar registry that is used
       to maintain a registry of OSPars that have been created
       on an ObjectStore.
    """
    @staticmethod
    def register(par, url_checksum, details_function, cleanup_function=None):
        """Register the passed PAR, passing in the checksum of
           the PAR's secret URL (so we can verify the close),
           and optionally supplying a cleanup_function that is
           called when the PAR is closed. The passed 'details_function'
           should be used to extract the object-store driver-specific
           details from the PAR and convert them into a dictionary.
           The signature should be;

           driver_details = details_function(par)
        """
        from Acquire.Service import is_running_service as _is_running_service

        if not _is_running_service():
            return

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        from Acquire.ObjectStore import OSPar as _OSPar

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import Function as _Function
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string

        if par is None:
            return

        if not isinstance(par, _OSPar):
            raise TypeError("You can only register pars of type PAR")

        if par.is_null():
            return

        data = {}
        data["par"] = par.to_data()

        if details_function is None:
            data["driver_details"] = par._driver_details
        else:
            data["driver_details"] = details_function(par)

        data["url_checksum"] = url_checksum

        if cleanup_function is not None:
            if not isinstance(cleanup_function, _Function):
                cleanup_function = _Function(cleanup_function)

            data["cleanup_function"] = cleanup_function.to_data()

        expire_string = _datetime_to_string(par.expires_when())

        key = "%s/uid/%s/%s" % (_registry_key, par.uid(), expire_string)

        bucket = _get_service_account_bucket()
        _ObjectStore.set_object_from_json(bucket, key, data)

        key = "%s/expire/%s/%s" % (_registry_key, expire_string, par.uid())
        _ObjectStore.set_object_from_json(bucket, key, par.uid())

    @staticmethod
    def get(par_uid, details_function, url_checksum=None):
        """Return the PAR that matches the passed PAR_UID.
           If 'url_checksum' is supplied then this verifies that
           the checksum of the secret URL is correct.

           This returns the PAR with a completed 'driver_details'.
           The 'driver_details' is created from the dictionary
           of data saved with the PAR. The signature should be;

           driver_details = details_function(data)
        """
        if par_uid is None or len(par_uid) == 0:
            return

        from Acquire.Service import is_running_service as _is_running_service

        if not _is_running_service():
            return

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        from Acquire.ObjectStore import OSPar as _OSPar

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import string_to_datetime \
            as _string_to_datetime

        key = "%s/uid/%s" % (_registry_key, par_uid)

        bucket = _get_service_account_bucket()

        objs = _ObjectStore.get_all_objects_from_json(bucket=bucket,
                                                      prefix=key)

        data = None

        for obj in objs.values():
            if url_checksum is not None:
                if url_checksum == obj["url_checksum"]:
                    data = obj
                    break
            else:
                data = obj
                break

        if data is None:
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "There is matching PAR available to close...")

        par = _OSPar.from_data(data["par"])

        if "driver_details" in data:
            if details_function is not None:
                driver_details = details_function(data["driver_details"])
                par._driver_details = driver_details
            else:
                par._driver_details = driver_details

        return par

    @staticmethod
    def close(par):
        """Close the passed PAR. This will remove the registration
           for the PAR and will also call the associated
           cleanup_function (if any)
        """
        from Acquire.Service import is_running_service as _is_running_service

        if not _is_running_service():
            return

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        from Acquire.ObjectStore import OSPar as _OSPar

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import Function as _Function

        if par is None:
            return

        if not isinstance(par, _OSPar):
            raise TypeError("You can only close OSPar objects!")

        if par.is_null():
            return

        expire_string = _datetime_to_string(par.expires_when())

        bucket = _get_service_account_bucket()

        key = "%s/expire/%s/%s" % (_registry_key, expire_string, par.uid())
        try:
            _ObjectStore.delete_object(bucket=bucket, key=key)
        except:
            pass

        key = "%s/uid/%s/%s" % (_registry_key, par.uid(), expire_string)

        try:
            data = _ObjectStore.take_object_from_json(bucket=bucket, key=key)
        except:
            data = None

        if data is None:
            # this PAR has already been closed
            return

        if "cleanup_function" in data:
            cleanup_function = _Function.from_data(data["cleanup_function"])
            cleanup_function(par=par)




__all__ = ["Drive"]


def _create_drive(metadata, creds):
    """Internal function used to create a Drive"""
    drive = Drive()
    drive._creds = creds

    from copy import copy as _copy
    metadata = _copy(metadata)
    metadata._set_credentials(creds)
    drive._metadata = metadata

    return drive


def _get_drive(creds, name=None, drive_uid=None,
               aclrules=None, autocreate=True):
    """Return the drive called 'name' using the passed credentials. The name
       will default to 'main' if it is not set, and the drive will
       be created automatically is 'autocreate' is True and the
       drive does not exist. If the drive is created, it would
       be created with the passed aclrules, if specified
       (this will be user-owner-only if not specified)
    """
    storage_service = creds.storage_service()

    if drive_uid is None:
        if name is None:
            name = "main"
        else:
            name = str(name)

        if autocreate:
            autocreate = True
        else:
            autocreate = False
    else:
        name = None
        autocreate = False
        drive_uid = str(drive_uid)

    args = {"name": name, "autocreate": autocreate,
            "drive_uid": drive_uid}

    if aclrules is not None:
        args["aclrules"] = aclrules.to_data()

    if creds.is_user():
        from Acquire.Client import Authorisation as _Authorisation
        authorisation = _Authorisation(resource="UserDrives",
                                       user=creds.user())
        args["authorisation"] = authorisation.to_data()
    elif creds.is_par():
        par = creds.par()
        par.assert_valid()
        args["par_uid"] = par.uid()
        args["secret"] = creds.secret()

    response = storage_service.call_function(function="open_drive", args=args)

    from Acquire.Client import DriveMeta as _DriveMeta

    return _create_drive(creds=creds,
                         metadata=_DriveMeta.from_data(response["drive"]))


class Drive:
    """This class provides a handle to a user's drive (space
       to hold files and folders). A drive is associated with
       a single storage service and can be shared amongst several
       users. Each drive has a unique UID, with users assiging
       their own shorthand names.

    """
    def __init__(self, name=None, drive_uid=None, creds=None,
                 aclrules=None, cheque=None, max_size=None,
                 autocreate=True):
        """Construct a handle to the drive that the passed user
           calls 'name' on the passed storage service. If
           'autocreate' is True and the user is logged in then
           this will automatically create the drive if
           it doesn't exist already
        """
        self._metadata = None
        self._creds = None

        if creds is not None:
            from Acquire.Client import StorageCreds as _StorageCreds
            if not isinstance(creds, _StorageCreds):
                raise TypeError("creds must be type StorageCreds")

            drive = _get_drive(creds=creds, name=name, drive_uid=drive_uid,
                               aclrules=aclrules, autocreate=autocreate)

            from copy import copy as _copy
            self.__dict__ = _copy(drive.__dict__)

    @staticmethod
    def open(metadata=None, drive_uid=None, creds=None):
        """Open and return the drive from the passed DriveMeta. The
           drive either needs to be opened via the User with
           storage service, or by passing in a valid PAR and secret
        """
        from Acquire.Client import DriveMeta as _DriveMeta
        from Acquire.Client import StorageCreds as _StorageCreds

        if not isinstance(metadata, _DriveMeta):
            raise TypeError("The metadata must be type DriveMeta")

        if metadata is None:
            return Drive(drive_uid=drive_uid, creds=creds)

        if creds is None:
            creds = metadata.credentials()

        if not isinstance(creds, _StorageCreds):
            raise TypeError("The creds must be type StorageCreds")

        return _create_drive(creds=creds, metadata=metadata)

    def __str__(self):
        if self.is_null():
            return "Drive::null"
        else:
            return "Drive(name='%s')" % self._metadata.name()

    def is_null(self):
        """Return whether or not this drive is null

           Returns:
                bool: True if null, else False
        """
        return self._metadata is None

    def metadata(self):
        """Return the metadata about this drive"""
        if self.is_null():
            return None
        else:
            from copy import copy as _copy
            return _copy(self._metadata)

    def credentials(self):
        """Return the credentials used to open this drive"""
        if self.is_null():
            return None
        else:
            return self._creds

    def storage_service(self):
        """Return the storage service for this drive
        Returns:
            Service: Storage service for drive
        """
        if self.is_null():
            return None
        else:
            return self._creds.storage_service()

    def chunk_upload(self, filename, dir=None, aclrules=None):
        """Start a chunked upload of a file called 'filename' (just the
           filename - not the full path - if you want to specify a certain
           directory in the Drive then specify that in 'dir').
           The file will be uploaded to the Drive at
           'dir/filename'. If a file with this name exists,
           then this will upload a new version (assuming you have permission).
           Otherwise this will create a new file. You can set the
           ACL rules used to grant access to this file via 'aclrules'.
           If this is not set, then the rules will be derived from either
           the last version of the file, or inherited from the drive.

           This will return a ChunkUploader which can be used to actually
           upload the file
        """
        if self.is_null():
            raise PermissionError("Cannot upload a file to a null drive!")

        if dir is not None:
            filename = "%s/%s" % (dir, filename)

        from Acquire.Client import FileMeta as _FileMeta
        filemeta = _FileMeta(filename=filename)
        filemeta._set_drive_metadata(self._metadata, self._creds)

        return filemeta.open().chunk_upload(aclrules=aclrules)

    def upload(self, filename, dir=None, uploaded_name=None, aclrules=None,
               force_par=False):
        """Upload the file at 'filename' to this drive, assuming we have
           write access to this drive (or all files in the directory
           at 'filename' if this is really a directory).
           The local file/directory 'filename' will be uploaded to the drive
           as the file/directory called 'filename' (just the
           filename - not the full path - if you want to specify a certain
           directory in the Drive then specify that in 'dir').
           If you want to specify the uploaded name then set this as
           "uploaded_name". The file/directory will be uploaded to the Drive at
           'dir/uploaded_name'. If a file with this name exists,
           then this will upload a new version (assuming you have permission).
           Otherwise this will create a new file. You can set the
           ACL rules used to grant access to this file via 'aclrules'.
           If this is not set, then the rules will be derived from either
           the last version of the file, or inherited from the drive.
        """
        if self.is_null():
            raise PermissionError("Cannot upload a file to a null drive!")

        if uploaded_name is None:
            import os as _os
            uploaded_name = _os.path.split(filename)[1]

        if dir is not None:
            uploaded_name = "%s/%s" % (dir, uploaded_name)

        import os as _os
        if _os.path.isdir(filename):
            # we need to upload each file in turn and then
            # return a location to a directory
            for f in _os.listdir(filename):
                self.upload(filename="%s/%s" % (filename, f),
                            uploaded_name="%s/%s" % (uploaded_name, f),
                            dir=None, aclrules=aclrules,
                            force_par=force_par)

            from Acquire.Client import DirMeta as _DirMeta
            dirmeta = _DirMeta(name=uploaded_name)
            dirmeta._set_drive_metadata(self._metadata, self._creds)

            return dirmeta
        else:
            from Acquire.Client import FileMeta as _FileMeta
            filemeta = _FileMeta(filename=uploaded_name)
            filemeta._set_drive_metadata(self._metadata, self._creds)

            return filemeta.open().upload(filename=filename,
                                          force_par=force_par,
                                          aclrules=aclrules)

    def chunk_download(self, filename, dir=None, download_name=None,
                       version=None):
        """Download the file 'filename' from the Drive to directory 'dir' on
           this computer (or current directory if not specified), calling
           the downloaded file 'download_filename' (or 'filename' if not
           specified). Force transfer using an OSPar is force_par is True
        """
        if self.is_null():
            raise PermissionError("Cannot upload a file to a null drive!")

        from Acquire.Client import FileMeta as _FileMeta
        filemeta = _FileMeta(filename=filename)
        filemeta._set_drive_metadata(self._metadata, self._creds)

        return filemeta.open().chunk_download(filename=download_name,
                                              version=version, dir=dir)

    def download(self, filename, dir=None, download_name=None,
                 version=None, force_par=False):
        """Download the file 'filename' from the Drive to directory 'dir' on
           this computer (or current directory if not specified), calling
           the downloaded file 'download_filename' (or 'filename' if not
           specified). Force transfer using an OSPar is force_par is True
        """
        if self.is_null():
            raise PermissionError("Cannot upload a file to a null drive!")

        from Acquire.Client import FileMeta as _FileMeta
        filemeta = _FileMeta(filename=filename)
        filemeta._set_drive_metadata(self._metadata, self._creds)

        return filemeta.open().download(filename=download_name,
                                        version=version, dir=dir,
                                        force_par=force_par)

    @staticmethod
    def _list_drives(creds, drive_uid=None):
        """Return a list of all of the DriveMetas of the drives accessible
           at the top-level using the passed credentials, or that are
           sub-drives of the drive with UID 'drive_uid'
        """
        from Acquire.Client import StorageCreds as _StorageCreds
        if not isinstance(creds, _StorageCreds):
            raise TypeError("The passed creds must be type StorageCreds")

        args = {}

        if creds.is_user():
            if drive_uid is not None:
                drive_uid = str(drive_uid)

            from Acquire.Client import Authorisation as _Authorisation
            authorisation = _Authorisation(resource="UserDrives",
                                           user=creds.user())

            args = {"authorisation": authorisation.to_data()}
        elif creds.is_par():
            par = creds.par()
            par.assert_valid()
            args["par"] = par.to_data()
            args["secret"] = creds.secret()

        if drive_uid is not None:
            args["drive_uid"] = str(drive_uid)

        storage_service = creds.storage_service()

        response = storage_service.call_function(
                                    function="list_drives", args=args)

        from Acquire.ObjectStore import string_to_list as _string_to_list
        from Acquire.Client import DriveMeta as _DriveMeta

        drives = _string_to_list(response["drives"], _DriveMeta)

        for drive in drives:
            drive._creds = creds

        return drives

    @staticmethod
    def list_toplevel_drives(creds):
        """Return a list of all of the DriveMetas of the drives accessible
           at the top-level using the passed credentils
        """
        return Drive._list_drives(creds=creds)

    def list_drives(self):
        """Return a list of the DriveMetas of all of the drives contained
           in this drive that are accessible to the user

           Returns:
                list: List of DriveMetas for the drives
        """
        if self.is_null():
            return []
        else:
            return Drive._list_drives(drive_uid=self._metadata.uid(),
                                      creds=self._creds)

    def list_files(self, dir=None, filename=None, include_metadata=False):
        """Return a list of the FileMetas of all of the files contained
           in this drive. If 'dir' is specified then list only the
           files that are contained in 'dir'. If 'filename' is specified
           then return only the files that match the passed filename
        """
        if self.is_null():
            return []

        from Acquire.ObjectStore import string_to_list as _string_to_list
        from Acquire.Storage import FileMeta as _FileMeta

        if include_metadata:
            include_metadata = True
        else:
            include_metadata = False

        args = {"drive_uid": self._metadata.uid(),
                "include_metadata": include_metadata}

        if dir is not None:
            args["dir"] = str(dir)

        if filename is not None:
            args["filename"] = str(filename)

        if self._creds.is_user():
            from Acquire.Client import Authorisation as _Authorisation
            authorisation = _Authorisation(resource="list_files",
                                           user=self._creds.user())
            args["authorisation"] = authorisation.to_data()
        elif self._creds.is_par():
            par = self._creds.par()
            par.assert_valid()
            args["par_uid"] = par.uid()
            args["secret"] = self._creds.secret()

        response = self.storage_service().call_function(function="list_files",
                                                        args=args)

        files = _string_to_list(response["files"], _FileMeta)

        for f in files:
            f._set_drive_metadata(self._metadata, self._creds)

        return files




import os as _os
import shutil as _shutil
import datetime as _datetime
import uuid as _uuid
import json as _json
import glob as _glob
import threading
import uuid as _uuid

_rlock = threading.RLock()

__all__ = ["Testing_ObjectStore"]


def _get_driver_details_from_par(par):
    from Acquire.ObjectStore import datetime_to_string \
        as _datetime_to_string

    import copy as _copy
    details = _copy.copy(par._driver_details)

    if details is None:
        return {}
    else:
        # fix any non-string/number objects
        details["created_datetime"] = _datetime_to_string(
                                        details["created_datetime"])

    return details


def _get_driver_details_from_data(data):
    from Acquire.ObjectStore import string_to_datetime \
        as _string_to_datetime

    import copy as _copy
    details = _copy.copy(data)

    if "created_datetime" in details:
        details["created_datetime"] = _string_to_datetime(
                                            details["created_datetime"])

    return details


class Testing_ObjectStore:
    """This is a dummy object store that writes objects to
       the standard posix filesystem when running tests
    """
    @staticmethod
    def create_bucket(bucket, bucket_name, compartment=None):
        """Create and return a new bucket in the object store called
           'bucket_name', optionally placing it into the compartment
           identified by 'compartment'. This will raise an
           ObjectStoreError if this bucket already exists
        """
        bucket_name = str(bucket_name)

        if compartment is not None:
            if compartment.endswith("/"):
                bucket = compartment
            else:
                bucket = "%s/" % compartment

        full_name = _os.path.join(_os.path.split(bucket)[0], bucket_name)

        if _os.path.exists(full_name):
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "CANNOT CREATE NEW BUCKET '%s': EXISTS!" % bucket_name)

        _os.makedirs(full_name)

        return full_name

    @staticmethod
    def get_bucket(bucket, bucket_name, compartment=None,
                   create_if_needed=True):
        """Find and return a new bucket in the object store called
           'bucket_name', optionally placing it into the compartment
           identified by 'compartment'. If 'create_if_needed' is True
           then the bucket will be created if it doesn't exist. Otherwise,
           if the bucket does not exist then an exception will be raised.
        """
        bucket_name = str(bucket_name)

        if compartment is not None:
            if compartment.endswith("/"):
                bucket = compartment
            else:
                bucket = "%s/" % compartment

        full_name = _os.path.join(_os.path.split(bucket)[0], bucket_name)

        if not _os.path.exists(full_name):
            if create_if_needed:
                _os.makedirs(full_name)
            else:
                from Acquire.ObjectStore import ObjectStoreError
                raise ObjectStoreError(
                    "There is no bucket available called '%s' in "
                    "compartment '%s'" % (bucket_name, compartment))

        return full_name

    @staticmethod
    def get_bucket_name(bucket):
        """Return the name of the passed bucket"""
        return _os.path.split(bucket)[1]

    @staticmethod
    def is_bucket_empty(bucket):
        """Return whether or not the passed bucket is empty"""
        return len(_os.listdir(bucket)) == 0

    @staticmethod
    def delete_bucket(bucket, force=False):
        """Delete the passed bucket. This should be used with caution.
           Normally you can only delete a bucket if it is empty. If
           'force' is True then it will remove all objects/pars from
           the bucket first, and then delete the bucket. This
           can cause a LOSS OF DATA!
        """
        is_empty = Testing_ObjectStore.is_bucket_empty(bucket=bucket)

        if not is_empty:
            if force:
                Testing_ObjectStore.delete_all_objects(bucket=bucket)
            else:
                raise PermissionError(
                    "You cannot delete the bucket %s as it is not empty" %
                    Testing_ObjectStore.get_bucket_name(bucket=bucket))

        # the bucket is empty - delete it
        _os.rmdir(bucket)

    @staticmethod
    def create_par(bucket, encrypt_key, key=None, readable=True,
                   writeable=False, duration=3600, cleanup_function=None):
        """Create a pre-authenticated request for the passed bucket and
           key (if key is None then the request is for the entire bucket).
           This will return a PAR object that will contain a URL that can
           be used to access the object/bucket. If writeable is true, then
           the URL will also allow the object/bucket to be written to.
           PARs are time-limited. Set the lifetime in seconds by passing
           in 'duration' (by default this is one hour). Note that you must
           pass in a public key that will be used to encrypt this PAR. This is
           necessary as the PAR grants access to anyone who can decrypt
           the URL
        """
        from Acquire.Crypto import PublicKey as _PublicKey

        if not isinstance(encrypt_key, _PublicKey):
            from Acquire.Client import PARError
            raise PARError(
                "You must supply a valid PublicKey to encrypt the "
                "returned PAR")

        if key is not None:
            if not _os.path.exists("%s/%s._data" % (bucket, key)):
                from Acquire.Client import PARError
                raise PARError(
                    "The object '%s' in bucket '%s' does not exist!" %
                    (key, bucket))
        elif not _os.path.exists(bucket):
            from Acquire.Client import PARError
            raise PARError("The bucket '%s' does not exist!" % bucket)

        url = "file://%s" % bucket

        if key:
            url = "%s/%s" % (url, key)

        # get the time this PAR was created
        from Acquire.ObjectStore import get_datetime_now as _get_datetime_now
        created_datetime = _get_datetime_now()

        # get the UTC datetime when this PAR should expire
        expires_datetime = created_datetime + \
            _datetime.timedelta(seconds=duration)

        # mimic limitations of OCI - cannot have a bucket PAR with
        # read permissions!
        if (key is None) and readable:
            from Acquire.Client import PARError
            raise PARError(
                "You cannot create a Bucket PAR that has read permissions "
                "due to a limitation in the underlying platform")

        from Acquire.ObjectStore import OSPar as _OSPar
        from Acquire.ObjectStore import OSParRegistry as _OSParRegistry

        url_checksum = _OSPar.checksum(url)

        driver_details = {"driver": "testing_objstore",
                          "bucket": bucket,
                          "created_datetime": created_datetime}

        par = _OSPar(url=url, key=key, encrypt_key=encrypt_key,
                     expires_datetime=expires_datetime,
                     is_readable=readable, is_writeable=writeable,
                     driver_details=driver_details)

        _OSParRegistry.register(par=par, url_checksum=url_checksum,
                                details_function=_get_driver_details_from_par,
                                cleanup_function=cleanup_function)

        return par

    @staticmethod
    def close_par(par=None, par_uid=None, url_checksum=None):
        """Close the passed PAR, which provides access to data in the
           passed bucket
        """
        from Acquire.ObjectStore import OSParRegistry as _OSParRegistry

        if par is None:
            par = _OSParRegistry.get(
                        par_uid=par_uid,
                        url_checksum=url_checksum,
                        details_function=_get_driver_details_from_data)

        from Acquire.ObjectStore import OSPar as _OSPar
        if not isinstance(par, _OSPar):
            raise TypeError("The PAR must be of type OSPar")

        if par.driver() != "testing_objstore":
            raise ValueError("Cannot delete a PAR that was not created "
                             "by the testing object store")

        # delete the PAR (no need to do this on testing)

        # close the PAR - this will trigger any close_function(s)
        _OSParRegistry.close(par=par)

    @staticmethod
    def get_object(bucket, key):
        """Return the binary data contained in the key 'key' in the
           passed bucket"""

        with _rlock:
            filepath = "%s/%s._data" % (bucket, key)
            if _os.path.exists(filepath):
                return open(filepath, "rb").read()
            else:
                from Acquire.ObjectStore import ObjectStoreError
                raise ObjectStoreError("No object at key '%s'" % key)

    @staticmethod
    def take_object(bucket, key):
        """Take (delete) the object from the object store, returning
           the object
        """
        with _rlock:
            filepath = "%s/%s._data" % (bucket, key)
            if _os.path.exists(filepath):
                data = open(filepath, "rb").read()
                _os.remove(filepath)
                return data
            else:
                from Acquire.ObjectStore import ObjectStoreError
                raise ObjectStoreError("No object at key '%s'" % key)

    @staticmethod
    def get_all_object_names(bucket, prefix=None, without_prefix=False):
        """Returns the names of all objects in the passed bucket"""

        root = bucket

        if prefix is not None:
            root = "%s/%s" % (bucket, prefix)

        root_len = len(bucket) + 1

        if without_prefix:
            prefix_len = len(prefix)

        subdir_names = _glob.glob("%s*" % root)

        object_names = []

        while True:
            names = subdir_names
            subdir_names = []

            for name in names:
                if name.endswith("._data"):
                    # remove the  ._data at the end
                    name = name[root_len:-6]
                    while name.endswith("/"):
                        name = name[0:-1]

                    if without_prefix:
                        name = name[prefix_len:]
                        while name.startswith("/"):
                            name = name[1:]

                    if len(name) > 0:
                        object_names.append(name)
                elif _os.path.isdir(name):
                    subdir_names += _glob.glob("%s/*" % name)

            if len(subdir_names) == 0:
                break

        return object_names

    @staticmethod
    def set_object(bucket, key, data):
        """Set the value of 'key' in 'bucket' to binary 'data'"""

        filename = "%s/%s._data" % (bucket, key)

        with _rlock:
            try:
                with open(filename, 'wb') as FILE:
                    if data is not None:
                        FILE.write(data)
                    FILE.flush()
            except:
                dir = "/".join(filename.split("/")[0:-1])
                _os.makedirs(dir, exist_ok=True)
                with open(filename, 'wb') as FILE:
                    if data is not None:
                        FILE.write(data)
                    FILE.flush()

    @staticmethod
    def delete_all_objects(bucket, prefix=None):
        """Deletes all objects..."""
        if prefix:
            _shutil.rmtree("%s/%s" % (bucket, prefix), ignore_errors=True)
        else:
            _shutil.rmtree(bucket, ignore_errors=True)

    @staticmethod
    def delete_object(bucket, key):
        """Removes the object at 'key'"""
        try:
            _os.remove("%s/%s._data" % (bucket, key))
        except:
            pass

    @staticmethod
    def get_size_and_checksum(bucket, key):
        """Return the object size (in bytes) and checksum of the
           object in the passed bucket at the specified key
        """
        filepath = "%s/%s._data" % (bucket, key)

        if not _os.path.exists(filepath):
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError("No object at key '%s'" % key)

        from Acquire.Access import get_filesize_and_checksum \
            as _get_filesize_and_checksum

        return _get_filesize_and_checksum(filepath)




__all__ = ["DriveInfo"]

_drive_root = "storage/drive"

_fileinfo_root = "storage/file"

_uploader_root = "storage/uploader"
_downloader_root = "storage/downloader"


def _validate_file_upload(par, file_bucket, file_key, objsize, checksum):
    """Call this function to signify that the file associated with
       this PAR has been uploaded. This will check that the
       objsize and checksum match with what was promised
    """
    from Acquire.ObjectStore import ObjectStore as _ObjectStore
    from Acquire.Service import get_service_account_bucket \
        as _get_service_account_bucket
    from Acquire.Service import get_this_service as _get_this_service

    service = _get_this_service()
    bucket = _get_service_account_bucket()

    file_bucket = _ObjectStore.get_bucket(
                        bucket=bucket, bucket_name=file_bucket,
                        compartment=service.storage_compartment(),
                        create_if_needed=True)

    # check that the file uploaded matches what was promised
    (real_objsize, real_checksum) = _ObjectStore.get_size_and_checksum(
                                                file_bucket, file_key)

    if real_objsize != objsize or real_checksum != checksum:
        # probably should delete the broken object here...

        from Acquire.Storage import FileValidationError
        raise FileValidationError(
            "The file uploaded does not match what was promised. "
            "size: %s versus %s, checksum: %s versus %s. Please try "
            "to upload the file again." %
            (real_objsize, objsize,
             real_checksum, checksum))

    # SHOULD HERE RECEIPT THE STORAGE TRANSACTION


class DriveInfo:
    """This class provides a service-side handle to the information
       about a particular cloud drive
    """
    def __init__(self, drive_uid=None, identifiers=None,
                 is_authorised=False, parent_drive_uid=None,
                 aclrules=None, autocreate=False):
        """Construct a DriveInfo for the drive with UID 'drive_uid',
           and optionally the GUID of the user making the request
           (and whether this was authorised). If this drive
           has a parent then it is a sub-drive and not recorded
           in the list of top-level drives
        """
        self._drive_uid = drive_uid
        self._parent_drive_uid = parent_drive_uid
        self._identifiers = identifiers
        self._is_authorised = is_authorised

        if self._drive_uid is not None:
            self.load(aclrules=aclrules, autocreate=autocreate)

    def __str__(self):
        if self.is_null():
            return "Drive::null"
        else:
            return "Drive(%s)" % self._drive_uid

    def is_null(self):
        return self._drive_uid is None

    def _drive_key(self):
        """Return the key for this drive in the object store"""
        return "%s/%s/info" % (_drive_root, self._drive_uid)

    def uid(self):
        """Return the UID of this drive"""
        return self._drive_uid

    def _get_metadata_bucket(self):
        """Return the bucket that contains all of the metadata about
           the files for this drive
        """
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.Service import get_this_service as _get_this_service

        service = _get_this_service()
        bucket = _get_service_account_bucket()
        bucket_name = "user_metadata"

        try:
            return _ObjectStore.get_bucket(
                            bucket=bucket, bucket_name=bucket_name,
                            compartment=service.storage_compartment(),
                            create_if_needed=True)
        except Exception as e:
            from Acquire.ObjectStore import RequestBucketError
            raise RequestBucketError(
                "Unable to open the bucket '%s': %s" % (bucket_name, str(e)))

    def _get_file_bucketname(self, filekey=None):
        """Return the name of the bucket that will contain all of the
           files for this drive.

           _filekey is passed in as a stub, for a future when we will
           need to split a drive over multiple object store buckets...
        """
        return "user_files"

    def _get_file_bucket(self, filekey=None):
        """Return the bucket that contains the file data for the
           file associated with 'filekey' in this drive
        """
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.Service import get_this_service as _get_this_service

        service = _get_this_service()
        bucket = _get_service_account_bucket()
        bucket_name = self._get_file_bucketname(filekey=filekey)

        try:
            return _ObjectStore.get_bucket(
                            bucket=bucket, bucket_name=bucket_name,
                            compartment=service.storage_compartment(),
                            create_if_needed=True)
        except Exception as e:
            from Acquire.ObjectStore import RequestBucketError
            raise RequestBucketError(
                "Unable to open the bucket '%s': %s" % (bucket_name, str(e)))

    def _resolve_acl(self, authorisation=None, resource=None,
                     accept_partial_match=False, par=None,
                     identifiers=None):
        """Internal function used to authorise access to this drive,
           returning the ACLRule of the access
        """
        if authorisation is not None:
            from Acquire.Client import Authorisation as _Authorisation
            if not isinstance(authorisation, _Authorisation):
                raise TypeError(
                    "The authorisation must be of type Authorisation")

            identifiers = authorisation.verify(
                            resource=resource,
                            accept_partial_match=accept_partial_match,
                            return_identifiers=True)
        elif par is not None:
            from Acquire.Client import PAR as _PAR
            if not isinstance(par, _PAR):
                raise TypeError("The par must be type PAR")

            if par.expired():
                raise PermissionError(
                    "Cannot access the drive at the PAR has expired!")

            if identifiers is None:
                raise PermissionError(
                    "The identifiers for the user who created the PAR "
                    "must be passed!")

            if par.location().drive_uid() != self._drive_uid:
                raise PermissionError(
                    "Cannot access the drive as the PAR is not "
                    "validated for this drive!")
        else:
            try:
                identifiers = self._identifiers
            except:
                identifiers = None

        try:
            upstream = self._upstream
        except:
            upstream = None

        drive_acl = self.aclrules().resolve(identifiers=identifiers,
                                            upstream=upstream,
                                            must_resolve=True,
                                            unresolved=False)

        if par is not None:
            drive_acl = drive_acl * par.aclrule()

        return (drive_acl, identifiers)

    def open_uploader(self, filename, aclrules=None, authorisation=None,
                      par=None, identifiers=None):
        """Create a return a ChunkUploader that will allow a file
           to be uploaded chunk-by-chunk (bit-by-bit). The filename
           of the files is passed as 'filename', and the aclrules
           specific for the file can also be set (otherwise will
           inherit from the drive). The authorisation, par and
           identifiers can be used to authenticate this request
        """
        (drive_acl, identifiers) = self._resolve_acl(
                        authorisation=authorisation,
                        resource="chunk_upload %s" % filename,
                        par=par, identifiers=identifiers)

        if not drive_acl.is_writeable():
            raise PermissionError(
                "You do not have permission to write to this drive. "
                "Your permissions are %s" % str(drive_acl))

        # now generate a FileInfo for this file
        from Acquire.Storage import FileInfo as _FileInfo
        fileinfo = _FileInfo(drive_uid=self._drive_uid,
                             filename=filename,
                             identifiers=identifiers,
                             upstream=drive_acl,
                             aclrules=aclrules,
                             is_chunked=True)

        # resolve the ACL for the file from this FileHandle
        filemeta = fileinfo.get_filemeta()
        file_acl = filemeta.acl()

        if not file_acl.is_writeable():
            raise PermissionError(
                "Despite having write permission to the drive, you "
                "do not have write permission for the file. Your file "
                "permissions are %s" % str(file_acl))

        from Acquire.Client import ChunkUploader as _ChunkUploader
        uploader = _ChunkUploader(drive_uid=self._drive_uid,
                                  file_uid=filemeta.uid())

        fileinfo.save()

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()
        key = "%s/%s/%s" % (_uploader_root, self._drive_uid, filemeta.uid())

        data = {"filename": filename,
                "version": filemeta.uid(),
                "filekey": fileinfo.latest_version()._file_key(),
                "secret": uploader.secret()}

        _ObjectStore.set_object_from_json(bucket, key, data)

        return (filemeta, uploader)

    def close_uploader(self, file_uid, secret):
        """Close the uploader associated with the passed file_uid,
           authenticated using the passed secret
        """
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()
        key = "%s/%s/%s" % (_uploader_root, self._drive_uid, file_uid)

        try:
            data = _ObjectStore.get_object_from_json(bucket, key)
        except:
            data = None

        if data is None:
            # the uploader has already been closed
            return

        shared_secret = data["secret"]

        if secret != shared_secret:
            raise PermissionError(
                "Invalid request - you do not have permission to "
                "close this uploader")

        try:
            data2 = _ObjectStore.take_object_from_json(bucket, key)
        except:
            data2 = None

        if data2 is None:
            # someone else is already in the process of closing
            # this uploader - let them do it!
            return

        filename = data["filename"]
        version = data["version"]

        # now get the FileInfo for this file
        from Acquire.Storage import FileInfo as _FileInfo
        fileinfo = _FileInfo.load(drive=self,
                                  filename=filename,
                                  version=version)

        file_key = data["filekey"]
        file_bucket = self._get_file_bucket(file_key)
        fileinfo.close_uploader(file_bucket=file_bucket)
        fileinfo.save()

    def close_downloader(self, downloader_uid, file_uid, secret):
        """Close the downloader associated with the passed
           downloader_uid and file_uid,
           authenticated using the passed secret
        """
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()
        key = "%s/%s/%s/%s" % (_downloader_root, self._drive_uid,
                               file_uid, downloader_uid)

        try:
            data = _ObjectStore.get_object_from_json(bucket, key)
        except:
            data = None

        if data is None:
            # the downloader has already been closed
            return

        shared_secret = data["secret"]

        if secret != shared_secret:
            raise PermissionError(
                "Invalid request - you do not have permission to "
                "close this downloader")

        try:
            _ObjectStore.take_object_from_json(bucket, key)
        except:
            pass

    def upload_chunk(self, file_uid, chunk_index, secret, chunk, checksum):
        """Upload a chunk of the file with UID 'file_uid'. This is the
           chunk at index 'chunk_idx', which is set equal to 'chunk'
           (validated with 'checksum'). The passed secret is used to
           authenticate this upload. The secret should be the
           multi_md5 has of the shared secret with the concatenated
           drive_uid, file_uid and chunk_index
        """
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()
        key = "%s/%s/%s" % (_uploader_root, self._drive_uid, file_uid)
        data = _ObjectStore.get_object_from_json(bucket, key)
        shared_secret = data["secret"]

        from Acquire.Crypto import Hash as _Hash
        shared_secret = _Hash.multi_md5(shared_secret,
                                        "%s%s%d" % (self._drive_uid,
                                                    file_uid,
                                                    chunk_index))

        if secret != shared_secret:
            raise PermissionError(
                "Invalid chunked upload secret. You do not have permission "
                "to upload chunks to this file!")

        # validate the data checksum
        check = _Hash.md5(chunk)

        if check != checksum:
            from Acquire.Storage import FileValidationError
            raise FileValidationError(
                "Invalid checksum for chunk: %s versus %s" %
                (check, checksum))

        meta = {"filesize": len(chunk),
                "checksum": checksum,
                "compression": "bz2"}

        file_key = data["filekey"]
        chunk_index = int(chunk_index)

        file_bucket = self._get_file_bucket(file_key)
        data_key = "%s/data/%d" % (file_key, chunk_index)
        meta_key = "%s/meta/%d" % (file_key, chunk_index)

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        _ObjectStore.set_object_from_json(file_bucket, meta_key, meta)
        _ObjectStore.set_object(file_bucket, data_key, chunk)

    def download_chunk(self, file_uid, downloader_uid, chunk_index, secret):
        """Download a chunk of the file with UID 'file_uid' at chunk
           index 'chunk_index'. This request is authenticated with
           the passed secret. The secret should be the
           multi_md5 has of the shared secret with the concatenated
           drive_uid, file_uid and chunk_index
        """
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()
        key = "%s/%s/%s/%s" % (_downloader_root, self._drive_uid,
                               file_uid, downloader_uid)

        try:
            data = _ObjectStore.get_object_from_json(bucket, key)
        except:
            data = None

        if data is None:
            raise PermissionError(
                "There is no downloader available to let you download "
                "this chunked file!")

        shared_secret = data["secret"]

        from Acquire.Crypto import Hash as _Hash
        shared_secret = _Hash.multi_md5(shared_secret,
                                        "%s%s%d" % (self._drive_uid,
                                                    file_uid,
                                                    chunk_index))

        if secret != shared_secret:
            raise PermissionError(
                "Invalid chunked upload secret. You do not have permission "
                "to upload chunks to this file!")

        file_key = data["filekey"]
        chunk_index = int(chunk_index)

        file_bucket = self._get_file_bucket(file_key)
        data_key = "%s/data/%d" % (file_key, chunk_index)
        meta_key = "%s/meta/%d" % (file_key, chunk_index)

        num_chunks = None

        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        try:
            meta = _ObjectStore.get_object_from_json(file_bucket, meta_key)
        except:
            meta = None

        if meta is None:
            # invalid read - see if the file has been closed?
            filename = data["filename"]
            version = data["version"]

            from Acquire.Storage import FileInfo as _FileInfo
            fileinfo = _FileInfo.load(drive=self,
                                      filename=filename,
                                      version=version)

            if fileinfo.version().is_uploading():
                raise IndexError("Invalid chunk index")

            num_chunks = fileinfo.version().num_chunks()

            if chunk_index < 0:
                chunk_index = num_chunks + chunk_index

            if chunk_index < 0 or chunk_index > num_chunks:
                raise IndexError("Invalid chunk index")
            elif chunk_index == num_chunks:
                # signal we've reached the end of the file
                return (None, None, num_chunks)

            # we should be able to read this metadata...
            meta = _ObjectStore.get_object_from_json(file_bucket, meta_key)

        chunk = _ObjectStore.get_object(file_bucket, data_key)

        return (chunk, meta, num_chunks)

    def upload(self, filehandle, authorisation=None, encrypt_key=None,
               par=None, identifiers=None):
        """Upload the file associated with the passed filehandle.
           If the filehandle has the data embedded, then this uploads
           the file data directly and returns a FileMeta for the
           result. Otherwise, this returns a PAR which should
           be used to upload the data. The PAR will be encrypted
           using 'encrypt_key'. Remember to close the PAR once the
           file has been uploaded, so that it can be validated
           as correct
        """
        from Acquire.Storage import FileHandle as _FileHandle
        from Acquire.Storage import FileInfo as _FileInfo
        from Acquire.Crypto import PublicKey as _PublicKey
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded

        if not isinstance(filehandle, _FileHandle):
            raise TypeError("The fileinfo must be of type FileInfo")

        if encrypt_key is not None:
            if not isinstance(encrypt_key, _PublicKey):
                raise TypeError("The encryption key must be of type PublicKey")

        (drive_acl, identifiers) = self._resolve_acl(
                        authorisation=authorisation,
                        resource="upload %s" % filehandle.fingerprint(),
                        par=par, identifiers=identifiers)

        if not drive_acl.is_writeable():
            raise PermissionError(
                "You do not have permission to write to this drive. "
                "Your permissions are %s" % str(drive_acl))

        # now generate a FileInfo for this FileHandle
        fileinfo = _FileInfo(drive_uid=self._drive_uid,
                             filehandle=filehandle,
                             identifiers=identifiers,
                             upstream=drive_acl)

        # resolve the ACL for the file from this FileHandle
        filemeta = fileinfo.get_filemeta()
        file_acl = filemeta.acl()

        if not file_acl.is_writeable():
            raise PermissionError(
                "Despite having write permission to the drive, you "
                "do not have write permission for the file. Your file "
                "permissions are %s" % str(file_acl))

        file_key = fileinfo.latest_version()._file_key()
        file_bucket = self._get_file_bucket(file_key)

        filedata = None

        if filehandle.is_localdata():
            # the filehandle already contains the file, so save it
            # directly
            filedata = filehandle.local_filedata()

        _ObjectStore.set_object(bucket=file_bucket,
                                key=file_key,
                                data=filedata)

        if filedata is None:
            # the file is too large to include in the filehandle so
            # we need to use a OSPar to upload
            from Acquire.ObjectStore import Function as _Function

            f = _Function(function=_validate_file_upload,
                          file_bucket=self._get_file_bucketname(),
                          file_key=file_key,
                          objsize=fileinfo.filesize(),
                          checksum=fileinfo.checksum())

            ospar = _ObjectStore.create_par(bucket=file_bucket,
                                            encrypt_key=encrypt_key,
                                            key=file_key,
                                            readable=False,
                                            writeable=True,
                                            cleanup_function=f)
        else:
            ospar = None

        # now save the fileinfo to the object store
        fileinfo.save()
        filemeta = fileinfo.get_filemeta()

        # return the PAR if we need to have a second-stage of upload
        return (filemeta, ospar)

    def download(self, filename, authorisation,
                 version=None, encrypt_key=None,
                 force_par=False, must_chunk=False,
                 par=None, identifiers=None):
        """Download the file called filename. This will return a
           FileHandle that describes the file. If the file is
           sufficiently small, then the filedata will be embedded
           into this handle. Otherwise a PAR will be generated and
           also returned to allow the file to be downloaded
           separately. The PAR will be encrypted with 'encrypt_key'.
           Remember to close the PAR once you have finished
           downloading the file...
        """
        from Acquire.Storage import FileHandle as _FileHandle
        from Acquire.Storage import FileInfo as _FileInfo
        from Acquire.Crypto import PublicKey as _PublicKey
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded

        if not isinstance(encrypt_key, _PublicKey):
            raise TypeError("The encryption key must be of type PublicKey")

        (drive_acl, identifiers) = self._resolve_acl(
                    authorisation=authorisation,
                    resource="download %s %s" % (self._drive_uid, filename),
                    par=par, identifiers=identifiers)

        # even if the drive_acl is not readable by this user, they
        # may have read permission for the file...

        # now get the FileInfo for this FileHandle
        fileinfo = _FileInfo.load(drive=self,
                                  filename=filename,
                                  version=version,
                                  identifiers=identifiers,
                                  upstream=drive_acl)

        # resolve the ACL for the file from this FileHandle
        filemeta = fileinfo.get_filemeta()
        file_acl = filemeta.acl()

        if not file_acl.is_readable():
            raise PermissionError(
                "You do not have read permissions for the file. Your file "
                "permissions are %s" % str(file_acl))

        file_bucket = self._get_file_bucket()

        file_key = fileinfo.version()._file_key()
        filedata = None
        downloader = None
        ospar = None

        if fileinfo.version().is_chunked():
            # this is a chunked file. We need to return a
            # ChunkDownloader to download the file
            from Acquire.Client import ChunkDownloader as _ChunkDownloader
            downloader = _ChunkDownloader(drive_uid=self._drive_uid,
                                          file_uid=fileinfo.version().uid())

            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket

            bucket = _get_service_account_bucket()
            key = "%s/%s/%s/%s" % (_downloader_root, self._drive_uid,
                                   filemeta.uid(), downloader.uid())

            data = {"filename": filename,
                    "version": filemeta.uid(),
                    "filekey": fileinfo.version()._file_key(),
                    "secret": downloader.secret()}

            _ObjectStore.set_object_from_json(bucket, key, data)

        elif must_chunk:
            raise PermissionError(
                "Cannot download this file in a chunked manner!")

        elif force_par or fileinfo.filesize() > 1048576:
            # the file is too large to include in the download so
            # we need to use a OSPar to download
            ospar = _ObjectStore.create_par(bucket=file_bucket,
                                            encrypt_key=encrypt_key,
                                            key=file_key,
                                            readable=True,
                                            writeable=False)
        else:
            # one-trip download of files that are less than 1 MB
            filedata = _ObjectStore.get_object(file_bucket, file_key)

        # return the filemeta, and either the filedata, ospar or downloader
        return (filemeta, filedata, ospar, downloader)

    def is_opened_by_owner(self):
        """Return whether or not this drive was opened and authorised
           by one of the drive owners
        """
        try:
            identifiers = self._identifiers
        except:
            identifiers = None

        if identifiers is None or (not self._is_authorised):
            return False

        try:
            upstream = self._upstream
        except:
            upstream = None

        drive_acl = self.aclrules().resolve(identifiers=identifiers,
                                            upstream=upstream)

        return drive_acl.is_owner()

    def aclrules(self):
        """Return the acl rules for this drive"""
        try:
            return self._aclrules
        except:
            from Acquire.Storage import ACLRules as _ACLRules
            return _ACLRules()

    def set_permission(self, user_guid, aclrule):
        """Set the permission for the user with the passed user_guid
           to 'aclrule". Note that you can only do this if you are the
           owner and this drive was opened in an authorised way. Also
           note that you cannot remove your own ownership permission
           if this would leave the drive without any owners
        """
        if self.is_null():
            return

        from Acquire.Identity import ACLRules as _ACLRules
        aclrules = _ACLRules.create(user_guid=user_guid, rule=aclrule)

        # make sure we have the latest version
        self.load(autocreate=False)

        if not self.is_opened_by_owner():
            raise PermissionError(
                "You cannot change user permissions as you are either "
                "not the owner of this drive or you failed to provide "
                "authorisation when you opened the drive")

        # this will append the new rules, ensuring that the change
        # does not leave the drive ownerless
        self._aclrules.append(aclrules, ensure_owner=True)

        self.save()
        self.load(autocreate=False)

    def list_files(self, authorisation=None, par=None,
                   identifiers=None, include_metadata=False,
                   dir=None, filename=None):
        """Return the list of FileMeta data for the files contained
           in this Drive. The passed authorisation is needed in case
           the list contents of this drive is not public.

           If 'dir' is specified, then only search for files in 'dir'.
           If 'filename' is specified, then only search for the
           file called 'filename'
        """
        (drive_acl, identifiers) = self._resolve_acl(
                                        authorisation=authorisation,
                                        resource="list_files",
                                        par=par, identifiers=identifiers)

        if par is not None:
            if par.location().is_file():
                dir = None
                filename = par.location().filename()
            elif not par.location().is_drive():
                raise PermissionError(
                    "You do not have permission to read the Drive")

        if not drive_acl.is_readable():
            raise PermissionError(
                "You don't have permission to read this Drive")

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import encoded_to_string as _encoded_to_string
        from Acquire.ObjectStore import string_to_encoded as _string_to_encoded
        from Acquire.Storage import FileMeta as _FileMeta

        metadata_bucket = self._get_metadata_bucket()

        if filename is not None:
            if dir is not None:
                filename = "%s/%s" % (dir, filename)

            key = "%s/%s/%s" % (_fileinfo_root, self._drive_uid,
                                _string_to_encoded(filename))

            names = [key]
        elif dir is not None:
            while dir.endswith("/"):
                dir = dir[0:-1]

            encoded_dir = _string_to_encoded(dir)

            while encoded_dir.endswith("="):
                encoded_dir = encoded_dir[0:-1]

            # remove the last two characters, as sometime uuencoding
            # will change the last characters so they don't match
            if len(encoded_dir) > 2:
                encoded_dir = encoded_dir[0:-2]
            else:
                encoded_dir = ""

            key = "%s/%s/%s" % (_fileinfo_root, self._drive_uid,
                                encoded_dir)

            all_names = _ObjectStore.get_all_object_names(metadata_bucket,
                                                          key)

            names = []

            dir = "%s/" % dir

            for name in all_names:
                decoded_name = _encoded_to_string(name.split("/")[-1])

                if decoded_name.startswith(dir):
                    names.append(name)
        else:
            key = "%s/%s" % (_fileinfo_root, self._drive_uid)
            names = _ObjectStore.get_all_object_names(metadata_bucket, key)

        files = []

        if include_metadata:
            # we need to load all of the metadata info for this file to
            # return to the user
            from Acquire.Storage import FileInfo as _FileInfo

            for name in names:
                try:
                    data = _ObjectStore.get_object_from_json(metadata_bucket,
                                                             name)
                    fileinfo = _FileInfo.from_data(data,
                                                   identifiers=identifiers,
                                                   upstream=drive_acl)
                    filemeta = fileinfo.get_filemeta()
                    file_acl = filemeta.acl()

                    if file_acl.is_readable() or file_acl.is_writeable():
                        files.append(filemeta)
                except:
                    pass
        else:
            for name in names:
                filename = _encoded_to_string(name.split("/")[-1])
                files.append(_FileMeta(filename=filename))

        return files

    def list_versions(self, filename, authorisation=None,
                      include_metadata=False, par=None,
                      identifiers=None):
        """Return the list of versions of the file with specified
           filename. If 'include_metadata' is true then this will
           load full metadata for each version. This will return
           a sorted list of FileMeta objects. The passed authorisation
           is needed in case the version info is not public
        """
        (drive_acl, identifiers) = self._resolve_acl(
                                    authorisation=authorisation,
                                    resource="list_versions %s" % filename,
                                    par=par, identifiers=identifiers)

        if not drive_acl.is_readable():
            raise PermissionError(
                "You don't have permission to read this Drive")

        from Acquire.Storage import FileInfo as _FileInfo
        versions = _FileInfo.list_versions(drive=self,
                                           filename=filename,
                                           identifiers=identifiers,
                                           upstream=drive_acl,
                                           include_metadata=include_metadata)

        result = []

        for version in versions:
            aclrules = version.aclrules()
            if aclrules is not None:
                acl = aclrules.resolve(upstream=drive_acl,
                                       identifiers=identifiers,
                                       must_resolve=True,
                                       unresolved=False)

                if acl.is_readable() or acl.is_writeable():
                    result.append(version)
            else:
                result.append(version)

        # return the versions sorted in upload order
        result.sort(key=lambda x: x.uploaded_when())

        return result

    def load(self, aclrules=None, autocreate=False):
        """Load the metadata about this drive from the object store"""
        if self.is_null():
            return

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        bucket = _get_service_account_bucket()

        drive_key = self._drive_key()

        try:
            data = _ObjectStore.get_object_from_json(bucket, drive_key)
        except:
            data = None

        if data is None:
            if not autocreate:
                raise PermissionError(
                    "There is no drive available with UID '%s'"
                    % self._drive_uid)

            # by default this user is the drive's owner
            try:
                user_guid = self._identifiers["user_guid"]
            except:
                user_guid = None

            if user_guid is None:
                # we cannot create the drive as we don't know who
                # requested it
                raise PermissionError(
                    "Cannot create the DriveInfo for a new drive as the "
                    "original request did not specify the creating user")
            elif not self._is_authorised:
                # we cannot create the drive if the request was not
                # authorised
                raise PermissionError(
                    "Cannot create the DriveInfo for a new drive as the "
                    "original request was not authorised by the user")

            # create a new drive and save it...
            from Acquire.Identity import ACLRule as _ACLRule
            from Acquire.Identity import ACLRules as _ACLRules

            if aclrules is not None:
                if not isinstance(aclrules, _ACLRules):
                    raise TypeError("The aclrules must be type ACLRules")

                aclrule = aclrules.resolve(
                            identifiers=self._identifiers)

                if not aclrule.is_owner():
                    raise PermissionError(
                        "You cannot set the permission of a new drive such "
                        "that you are not the owner! aclrules = %s, resolved "
                        "aclrule for identifiers = %s is %s" %
                        (aclrules, self._identifiers, aclrule))

                self._aclrules = aclrules
            else:
                self._aclrules = _ACLRules.owner(user_guid=user_guid)

            data = self.to_data()

            data = _ObjectStore.set_ins_object_from_json(bucket, drive_key,
                                                         data)

        from copy import copy as _copy
        other = DriveInfo.from_data(data)

        identifiers = self._identifiers
        is_authorised = self._is_authorised

        self.__dict__ = _copy(other.__dict__)

        self._identifiers = identifiers
        self._is_authorised = is_authorised

    def save(self):
        """Save the metadata about this drive to the object store"""
        if self.is_null():
            return

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        bucket = _get_service_account_bucket()

        drive_key = self._drive_key()

        data = self.to_data()
        _ObjectStore.set_object_from_json(bucket, drive_key, data)

    def to_data(self):
        """Return a json-serialisable dictionary for this object"""
        data = {}

        if not self.is_null():
            from Acquire.ObjectStore import dict_to_string as _dict_to_string
            data["uid"] = self._drive_uid

            if self._aclrules is not None:
                data["aclrules"] = self._aclrules.to_data()

        return data

    @staticmethod
    def from_data(data):
        """Return a DriveInfo constructed from the passed json-deserialised
           dictionary
        """
        info = DriveInfo()

        if data is None or len(data) == 0:
            return info

        info._drive_uid = str(data["uid"])

        if "aclrules" in data:
            from Acquire.Storage import ACLRules as _ACLRules
            info._aclrules = _ACLRules.from_data(data["aclrules"])

        return info




__all__ = ["UserCredentials"]

_user_root = "identity/users"


class UserCredentials:
    """This class is used to store user credentials in the object
       store, and to verify that user credentials are correct.

       The user credentials are used to ultimately store a
       primary password for the user, which unlocks the user's
       primary private key
    """
    @staticmethod
    def hash(username, password, service_uid=None):
        """Return a secure hash of the passed username and password"""
        from Acquire.Crypto import Hash as _Hash
        from Acquire.Service import get_this_service as _get_this_service

        if service_uid is None:
            service_uid = _get_this_service(need_private_access=False).uid()

        result = _Hash.multi_md5(service_uid, username+password)

        return result

    @staticmethod
    def create(user_uid, password, primary_password,
               device_uid=None):
        """Create the credentials for the user with specified
           user_uid, optionally logging in via the specified
           device_uid, using the specified password, to protect
           the passed "primary_password"

           This returns the OTP that has been created to be
           associated with these credentials
        """
        from Acquire.Crypto import PrivateKey as _PrivateKey
        from Acquire.Crypto import OTP as _OTP
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.ObjectStore import bytes_to_string as _bytes_to_string

        if device_uid is None:
            device_uid = user_uid

        privkey = _PrivateKey(name="user_creds_key %s" % user_uid)
        otp = _OTP()
        otpsecret = otp.encrypt(privkey.public_key())
        primary_password = privkey.encrypt(primary_password)

        data = {"primary_password": _bytes_to_string(primary_password),
                "private_key": privkey.to_data(passphrase=password),
                "otpsecret": _bytes_to_string(otpsecret)
                }

        key = "%s/credentials/%s/%s" % (_user_root, user_uid, device_uid)

        bucket = _get_service_account_bucket()
        _ObjectStore.set_object_from_json(bucket=bucket,
                                          key=key,
                                          data=data)

        return otp

    @staticmethod
    def login(credentials, user_uids, remember_device=False):
        """Verify the passed credentials are correct.
           This will find the account
           that matches these credentials. If one does, then this
           will validate the credentials are correct, and then return
           a tuple of the (user_uid, primary_password) for that
           user
        """
        from Acquire.Crypto import PrivateKey as _PrivateKey
        from Acquire.Crypto import OTP as _OTP
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.ObjectStore import string_to_bytes as _string_to_bytes
        from Acquire.Client import Credentials as _Credentials
        from Acquire.Crypto import RepeatedOTPCodeError \
            as _RepeatedOTPCodeError

        if not isinstance(credentials, _Credentials):
            raise TypeError("The passed credentials must be type Credentials")

        username = credentials.username()
        short_uid = credentials.short_uid()
        device_uid = credentials.device_uid()
        password = credentials.password()
        otpcode = credentials.otpcode()

        bucket = _get_service_account_bucket()

        # now try to find a matching user_uid
        for user_uid in user_uids:
            data = None

            try:
                key = "%s/credentials/%s/%s" % (_user_root, user_uid,
                                                device_uid)
                data = _ObjectStore.get_object_from_json(bucket=bucket,
                                                         key=key)
                verified_device_uid = device_uid
            except:
                pass

            if data is None:
                # unknown device_uid. Try using the user_uid
                key = "%s/credentials/%s/%s" % (_user_root, user_uid,
                                                user_uid)
                try:
                    data = _ObjectStore.get_object_from_json(bucket=bucket,
                                                             key=key)
                    verified_device_uid = None
                except:
                    pass

            if data is not None:
                # verify the credentials
                try:
                    return UserCredentials.validate_password(
                                            user_uid=user_uid,
                                            username=username,
                                            device_uid=verified_device_uid,
                                            secrets=data,
                                            password=password,
                                            otpcode=otpcode,
                                            remember_device=remember_device)
                except _RepeatedOTPCodeError as e:
                    # if the OTP code is entered twice, then we need
                    # to invalidate the other session
                    raise e
                except:
                    # this is not the matching user...
                    pass

        # only get here if there are no matching users (or the
        # user-supplied password etc. are wrong)
        from Acquire.Identity import UserValidationError
        raise UserValidationError(
            "Invalid credentials logging into session '%s' "
            "with username '%s'" % (short_uid, username))

    @staticmethod
    def validate_password(user_uid, username, device_uid, secrets, password,
                          otpcode, remember_device):
        """Validate that the passed password and one-time-code are valid.
           If they are, then return a tuple of the UserAccount of the unlocked
           user, the OTP that is used to generate secrets, and the
           device_uid of the login device

           If 'remember_device' is True and 'device_uid' is None, then
           this creates a new OTP for the login device, which is returned,
           and a new device_uid for that device. The password needed to
           match this device is a MD5 of the normal user password.
        """
        from Acquire.Crypto import PrivateKey as _PrivateKey
        from Acquire.Crypto import OTP as _OTP
        from Acquire.ObjectStore import string_to_bytes as _string_to_bytes

        privkey = _PrivateKey.from_data(data=secrets["private_key"],
                                        passphrase=password)

        # decrypt and validate the OTP code
        data = _string_to_bytes(secrets["otpsecret"])

        otpsecret = privkey.decrypt(data)
        otp = _OTP(secret=otpsecret)
        otp.verify(code=otpcode, once_only=True)

        # everything is ok - we can load the user account via the
        # decrypted primary password
        primary_password = _string_to_bytes(secrets["primary_password"])
        primary_password = privkey.decrypt(primary_password)

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        data = None
        secrets = None
        key = "%s/uids/%s" % (_user_root, user_uid)

        bucket = _get_service_account_bucket()

        try:
            data = _ObjectStore.get_object_from_json(bucket=bucket, key=key)
        except:
            pass

        if data is None:
            from Acquire.Identity import UserValidationError
            raise UserValidationError(
                "Unable to validate user as no account data is present!")

        from Acquire.Identity import UserAccount as _UserAccount
        user = _UserAccount.from_data(data=data, passphrase=primary_password)

        if user.uid() != user_uid:
            from Acquire.Identity import UserValidationError
            raise UserValidationError(
                "Unable to validate user as mismatch in user_uids!")

        if device_uid is None and remember_device:
            # create a new OTP that is unique for this device
            from Acquire.ObjectStore import create_uuid as _create_uuid
            from Acquire.Client import Credentials as _Credentials
            device_uid = _create_uuid()
            device_password = _Credentials.encode_device_uid(
                                                encoded_password=password,
                                                device_uid=device_uid)

            otp = UserCredentials.create(user_uid=user_uid,
                                         password=device_password,
                                         primary_password=primary_password,
                                         device_uid=device_uid)

            # now save a lookup so that we can find the user_uid from
            # the username and device-specific password
            encoded_password = UserCredentials.hash(
                                        username=username,
                                        password=device_password)

            key = "%s/passwords/%s/%s" % (_user_root, encoded_password,
                                          user_uid)

            from Acquire.ObjectStore import get_datetime_now_to_string \
                as _get_datetime_now_to_string

            _ObjectStore.set_string_object(
                                bucket=bucket, key=key,
                                string_data=_get_datetime_now_to_string())

        return {"user": user, "otp": otp, "device_uid": device_uid}




# use a variable so we can monkey-patch while testing
_input = input

__all__ = ["Wallet"]


def _get_wallet_dir():
    """Function that can be mocked in testing to ensure we don't
       mess with the user's main wallet
    """
    import os as _os
    home = _os.path.expanduser("~")
    return "%s/.acquire_wallet" % home


def _get_wallet_password(confirm_password=False):
    """Function that can be mocked in testing to ensure we don't
       mess with the user's main wallet
    """
    import getpass as _getpass
    password = _getpass.getpass(
                prompt="Please enter a password to encrypt your wallet: ")

    if not confirm_password:
        return password

    password2 = _getpass.getpass(
                    prompt="Please confirm the password: ")

    if password != password2:
        _output("The passwords don't match. Please try again.")
        return _get_wallet_password(confirm_password=confirm_password)

    return password


def _output(s, end=None):
    """Simple output function that can be replaced at testing
       to silence the wallet
    """
    if end is None:
        print(s)
    else:
        print(s, end=end)


def _flush_output():
    """Flush STDOUT"""
    try:
        import sys as _sys
        _sys.stdout.flush()
    except:
        pass


def _read_json(filename):
    """Return a json-decoded dictionary from the data written
       to 'filename'
    """
    import json as _json
    with open(filename, "rb") as FILE:
        s = FILE.read().decode("utf-8")

        return _json.loads(s)


def _write_json(data, filename):
    """Write the passed json-encodable dictionary to 'filename'"""
    import json as _json
    s = _json.dumps(data)
    with open(filename, "wb") as FILE:
        FILE.write(s.encode("utf-8"))


def _read_service(filename):
    """Read and return the service written to 'filename'"""
    from Acquire.Client import Service as _Service
    return _Service.from_data(_read_json(filename))


def _write_service(service, filename):
    """Write the passed service to 'filename'"""
    _write_json(service.to_data(), filename)


def _could_match(userinfo, username, password):
    if username is None:
        return True

    if "username" not in userinfo:
        return False

    if userinfo["username"] == username:
        if password is None:
            return True

        if "password" in userinfo:
            if userinfo["password"] == password:
                return True

    return False


class Wallet:
    """This class holds a wallet that can be used to simplify
       sending passwords and one-time-password (OTP) codes
       to an acquire identity service.

       This holds a wallet of passwords and (optionally)
       OTP secrets that are encrypted using a local keypair
       that is unlocked by a password supplied by the user locally.

       By default this will create the wallet in your home
       directory ($HOME/.acquire_wallet). If you want the wallet
       to be saved in a different directory, specify that
       as "wallet_dir".
    """
    def __init__(self):
        """Construct a wallet to hold all user credentials"""
        from Acquire.Service import is_running_service \
            as _is_running_service

        if _is_running_service():
            from Acquire.Service import get_this_service \
                as _get_this_service
            service = _get_this_service(need_private_access=False)
            raise PermissionError(
                "You cannot open a Wallet on a running Service (%s)" %
                service)

        self._wallet_key = None
        self._service_info = {}

        import os as _os

        wallet_dir = _get_wallet_dir()

        if not _os.path.exists(wallet_dir):
            _os.makedirs(wallet_dir, mode=0o700, exist_ok=False)
        elif not _os.path.isdir(wallet_dir):
            raise TypeError("The wallet directory must be a directory!")

        self._wallet_dir = wallet_dir

    def _create_wallet_key(self, filename):
        """Create a new wallet key for the user"""
        password = _get_wallet_password(confirm_password=True)

        from Acquire.Client import PrivateKey as _PrivateKey
        key = _PrivateKey(name="wallet_key")

        bytes = key.bytes(password)

        with open(filename, "wb") as FILE:
            FILE.write(bytes)

        return key

    def _get_wallet_key(self):
        """Return the private key used to encrypt everything in the wallet.
           This will ask for the users password
        """
        if self._wallet_key:
            return self._wallet_key

        wallet_dir = self._wallet_dir

        keyfile = "%s/wallet_key.pem" % wallet_dir

        import os as _os

        if not _os.path.exists(keyfile):
            self._wallet_key = self._create_wallet_key(filename=keyfile)
            return self._wallet_key

        # read the keyfile and decrypt
        with open(keyfile, "rb") as FILE:
            bytes = FILE.read()

        wallet_key = None

        from Acquire.Client import PrivateKey as _PrivateKey

        # get the user password
        for _ in range(0, 5):
            password = _get_wallet_password()

            try:
                wallet_key = _PrivateKey.read_bytes(bytes, password)
            except:
                _output("Invalid password. Please try again.")

            if wallet_key:
                break

        if wallet_key is None:
            raise PermissionError("Too many failed password attempts...")

        self._wallet_key = wallet_key
        return wallet_key

    def _get_userinfo_filename(self, user_uid, identity_uid):
        """Return the filename for the passed user_uid logging into the
           passed identity service with identity_uid
        """
        assert(user_uid is not None)
        assert(identity_uid is not None)

        return "%s/user_%s_%s_encrypted" % (
            self._wallet_dir, user_uid, identity_uid)

    def _set_userinfo(self, userinfo, user_uid, identity_uid):
        """Save the userfile for the passed user_uid logging into the
           passed identity service with identity_uid
        """
        from Acquire.ObjectStore import bytes_to_string as _bytes_to_string
        import json as _json
        filename = self._get_userinfo_filename(user_uid=user_uid,
                                               identity_uid=identity_uid)
        key = self._get_wallet_key().public_key()
        data = _bytes_to_string(key.encrypt(_json.dumps(userinfo)))

        userinfo = {"username": userinfo["username"],
                    "user_uid": user_uid,
                    "data": data}

        _write_json(data=userinfo, filename=filename)

    def _unlock_userinfo(self, userinfo):
        """Function used to unlock (decrypt) the passed userinfo"""
        from Acquire.ObjectStore import string_to_bytes as _string_to_bytes
        import json as _json

        key = self._get_wallet_key()

        data = _string_to_bytes(userinfo["data"])
        result = _json.loads(key.decrypt(data))
        result["user_uid"] = userinfo["user_uid"]
        return result

    def _get_userinfo(self, user_uid, identity_uid):
        """Read all info for the passed user at the identity service
           reached at 'identity_url'
        """
        filename = self._get_userinfo_filename(user_uid=user_uid,
                                               identity_uid=identity_uid)
        userinfo = _read_json(filename=filename)
        return self._unlock_userinfo(userinfo)

    def _find_userinfo(self, username=None, password=None):
        """Function to find a user_info automatically, of if that fails,
           to ask the user
        """
        wallet_dir = self._wallet_dir

        import glob as _glob

        userfiles = _glob.glob("%s/user_*_encrypted" % wallet_dir)

        userinfos = []

        for userfile in userfiles:
            try:
                userinfo = _read_json(userfile)
                if _could_match(userinfo, username, password):
                    userinfos.append((userinfo["username"], userinfo))
            except:
                pass

        userinfos.sort(key=lambda x: x[0])

        if len(userinfos) == 1:
            return self._unlock_userinfo(userinfos[0][1])

        if len(userinfos) == 0:
            if username is None:
                username = _input("Please type your username: ")

            userinfo = {"username": username}

            if password is not None:
                userinfo["password"] = password

            return userinfo

        _output("Please choose the account by typing in its number, "
                "or type a new username if you want a different account.")

        for (i, (username, userinfo)) in enumerate(userinfos):
            _output("[%d] %s {%s}" % (i+1, username, userinfo["user_uid"]))

        max_tries = 5

        for i in range(0, max_tries):
            reply = _input(
                    "\nMake your selection (1 to %d) " %
                    (len(userinfos))
                )

            try:
                idx = int(reply) - 1
            except:
                idx = None

            if idx is None:
                # interpret this as a username
                return self._find_userinfo(username=reply, password=password)
            elif idx < 0 or idx >= len(userinfos):
                _output("Invalid account.")
            else:
                return self._unlock_userinfo(userinfos[idx][1])

            if i < max_tries-1:
                _output("Try again...")

        userinfo = {}

        if username is not None:
            userinfo["username"] = username

        return userinfo

    def _get_user_password(self, userinfo):
        """Get the user password for the passed user on the passed
           identity_url
        """
        if "password" in userinfo:
            return userinfo["password"]
        else:
            import getpass as _getpass
            password = _getpass.getpass(
                            prompt="Please enter the login password: ")
            userinfo["password"] = password
            return password

    def _get_otpcode(self, userinfo):
        """Get the OTP code"""
        if "otpsecret" in userinfo:
            from Acquire.Client import OTP as _OTP
            otp = _OTP(userinfo["otpsecret"])
            return otp.generate()
        else:
            import getpass as _getpass
            return _getpass.getpass(
                        prompt="Please enter the one-time-password code: ")

    def get_services(self):
        """Return all of the trusted services known to this wallet"""
        import glob as _glob
        service_files = _glob.glob("%s/service_*" % self._wallet_dir)

        services = []

        for service_file in service_files:
            services.append(_read_service(service_file))

        return services

    def get_service(self, service_url=None, service_uid=None,
                    service_type=None, autofetch=True):
        """Return the service at either 'service_url', or that
           has UID 'service_uid'. This will return the
           cached service if it exists, or will add a new service if
           we are able to validate it from a trusted registry
        """
        from Acquire.ObjectStore import string_to_safestring \
            as _string_to_safestring

        service = None

        if service_url is None:
            if service_uid is None:
                raise PermissionError(
                    "You need to specify one of service_uid or service_url")

            # we need to look up the name...
            import glob as _glob
            service_files = _glob.glob("%s/service_*" % self._wallet_dir)

            for service_file in service_files:
                s = _read_service(service_file)
                if s.uid() == service_uid:
                    service = s
                    break
        else:
            from Acquire.Service import Service as _Service
            service_url = _Service.get_canonical_url(service_url,
                                                     service_type=service_type)

            service_file = "%s/service_%s" % (
                self._wallet_dir,
                _string_to_safestring(service_url))

            try:
                service = _read_service(service_file)
            except:
                pass

        must_write = False

        if service is None:
            if not autofetch:
                from Acquire.Service import ServiceError
                raise ServiceError("No service at %s:%s" %
                                   (service_url, service_uid))

            # we need to look this service up from the registry
            from Acquire.Registry import get_trusted_registry_service \
                as _get_trusted_registry_service

            _output("Connecting to registry...")
            _flush_output()

            registry = _get_trusted_registry_service(service_uid=service_uid,
                                                     service_url=service_url)

            _output("...connected to registry %s" % registry)
            _flush_output()

            # ensure we cache this registry...
            registry_file = "%s/service_%s" % (
                self._wallet_dir,
                _string_to_safestring(registry.canonical_url()))
            _write_service(service=registry, filename=registry_file)

            if service_url is not None:
                _output("Securely fetching keys for %s..." % service_url)
                _flush_output()
            else:
                _output("Securely fetching keys for UID %s..." % service_uid)
                _flush_output()

            service = registry.get_service(service_url=service_url,
                                           service_uid=service_uid)

            _output("...success.\nFetched %s" % service)
            _flush_output()

            must_write = True

        # check if the keys need rotating - if they do, load up
        # the new keys and save them to the service file...
        if service.should_refresh_keys():
            service.refresh_keys()
            must_write = True

        if service_uid is not None:
            if service.uid() != service_uid:
                raise PermissionError(
                    "Disagreement over the service UID for '%s' (%s)" %
                    (service, service_uid))

        if must_write:
            service_file = "%s/service_%s" % (
                self._wallet_dir,
                _string_to_safestring(service.canonical_url()))
            _write_service(service=service, filename=service_file)

        return service

    def remove_all_services(self):
        """Remove all trusted services from this Wallet"""
        import glob as _glob
        import os as _os
        service_files = _glob.glob("%s/service_*" % self._wallet_dir)

        for service_file in service_files:
            if _os.path.exists(service_file):
                _os.unlink(service_file)

    def remove_service(self, service):
        """Remove the cached service info for the passed service"""
        if isinstance(service, str):
            service_url = service
        else:
            service_url = service.canonical_url()

        from Acquire.ObjectStore import string_to_safestring \
            as _string_to_safestring

        service_file = "%s/service_%s" % (
            self._wallet_dir,
            _string_to_safestring(service_url))

        import os as _os

        if _os.path.exists(service_file):
            _os.unlink(service_file)

    def send_password(self, url, username=None, password=None,
                      otpcode=None, remember_password=True,
                      remember_device=None, dryrun=None):
        """Send a password and one-time code to the supplied login url"""
        if not remember_password:
            remember_device = False

        # the login URL is http[s]://something.com?id=XXXX/YY.YY.YY.YY
        # where XXXX is the service_uid of the service we should
        # connect with, and YY.YY.YY.YY is the short_uid of the login
        try:
            from urllib.parse import urlparse as _urlparse
            from urllib.parse import parse_qs as _parse_qs
            idcode = _parse_qs(_urlparse(url).query)["id"][0]
        except Exception as e:
            from Acquire.Client import LoginError
            raise LoginError(
                "Cannot identify the session or service information from "
                "the login URL '%s'. This should have id=XX-XX/YY.YY.YY.YY "
                "as a query parameter. <%s> %s" %
                (url, e.__class__.__name__, str(e)))

        try:
            (service_uid, short_uid) = idcode.split("/")
        except:
            from Acquire.Client import LoginError
            raise LoginError(
                "Cannot extract the service_uid and short_uid from the "
                "login ID code '%s'. This should be in the format "
                "XX-XX/YY.YY.YY.YY" % idcode)

        # now get the service
        try:
            service = self.get_service(service_uid=service_uid)
        except Exception as e:
            from Acquire.Client import LoginError
            raise LoginError(
                "Cannot find the service with UID %s: <%s> %s" %
                (service_uid, e.__class__.__name__, str(e)))

        if not service.can_identify_users():
            from Acquire.Client import LoginError
            raise LoginError(
                "Service '%s' is unable to identify users! "
                "You cannot log into something that is not "
                "a valid identity service!" % (service))

        userinfo = self._find_userinfo(username=username,
                                       password=password)

        if username is None:
            username = userinfo["username"]

        if "user_uid" in userinfo:
            user_uid = userinfo["user_uid"]
        else:
            user_uid = None

        _output("Logging in using username '%s'" % username)

        try:
            device_uid = userinfo["device_uid"]
        except:
            device_uid = None

        if password is None:
            password = self._get_user_password(userinfo=userinfo)

        if otpcode is None:
            otpcode = self._get_otpcode(userinfo=userinfo)
        else:
            # user is providing the primary OTP, so this is not a device
            device_uid = None

        _output("\nLogging in to '%s', session '%s'..." % (
                service.canonical_url(), short_uid), end="")

        _flush_output()

        if dryrun:
            print("Calling %s with username=%s, password=%s, otpcode=%s, "
                  "remember_device=%s, device_uid=%s, short_uid=%s "
                  "user_uid=%s" %
                  (service.canonical_url(), username, password, otpcode,
                   remember_device, device_uid, short_uid, user_uid))
            return

        try:
            from Acquire.Client import Credentials as _Credentials

            creds = _Credentials(username=username, password=password,
                                 otpcode=otpcode, short_uid=short_uid,
                                 device_uid=device_uid)

            args = {"credentials": creds.to_data(identity_uid=service.uid()),
                    "user_uid": user_uid,
                    "remember_device": remember_device,
                    "short_uid": short_uid}

            response = service.call_function(function="login", args=args)
            _output("SUCCEEDED!")
            _flush_output()
        except Exception as e:
            _output("FAILED!")
            _flush_output()
            from Acquire.Client import LoginError
            raise LoginError("Failed to log in. %s" % e.args)

        if not remember_password:
            return

        try:
            returned_user_uid = response["user_uid"]

            if returned_user_uid != user_uid:
                # change of user?
                userinfo = {}
                user_uid = returned_user_uid
        except:
            # no user_uid, so nothing to save
            return

        if user_uid is None:
            # can't save anything
            return

        userinfo["username"] = username
        userinfo["password"] = password

        try:
            userinfo["device_uid"] = response["device_uid"]
        except:
            pass

        try:
            userinfo["otpsecret"] = response["otpsecret"]
        except:
            pass

        self._set_userinfo(userinfo=userinfo,
                           user_uid=user_uid, identity_uid=service.uid())




__all__ = ["FileHandle"]


_magic_dict = {
    b"\x1f\x8b\x08": "gz",
    b"\x42\x5a\x68": "bz2",
    b"\x50\x4b\x03\x04": "zip"
    }


_max_magic_len = max(len(x) for x in _magic_dict)


def _should_compress(filename, filesize):
    """Return whether or not the passed file is worth compressing.
       It is not worth compressing very small files (<128 bytes) or
       already-compressed files

       Args:
            filename (str): Filename
            filesize (int): Size of file in bytes
       Returns:
            bool: True if file should be compressed, else
            False
    """
    if filesize < 128:
        return False

    with open(filename, "rb") as FILE:
        file_start = FILE.read(_max_magic_len)

    for magic in _magic_dict.keys():
        if file_start.startswith(magic):
            return False

    return True


def _bz2compress(inputfile, outputfile=None):
    """Compress 'inputfile', writing the output to 'outputfile'
       If 'outputfile' is None, then this will create a new filename
       in the current directory for the file. This returns
       the filename for the compressed file

       Args:
            inputfile (str): File to compress
            outputfile (str, default=None): Name for compressed
            file
       Returns:
            str: Filename of compressed file

    """
    from Acquire.Client import compress as _compress
    return _compress(inputfile=inputfile, outputfile=outputfile,
                     compression_type="bz2")


class FileHandle:
    """This class holds all of the information about a file that is
       held in a Drive, including its size
       and checksum, and information about previous versions. It
       provides a handle that you can use to download or delete
       the file, to upload new versions, or to move the data between
       hot and cold storage or pay for extended storage

       Args:
            filename (str, default=None): File to create handle for
            remote_filename (str, default=None): Remote path for file
            aclrules (str, default=None): ACL rules for handle
            drive_uid (str, default=None): UID for drive
            compress (bool, default=True): Should files be compressed
            local_cutoff (int, default=None): Size of file to be held
            locally by the handle (bytes)

    """
    def __init__(self, filename=None, remote_filename=None,
                 aclrules=None, drive_uid=None,
                 compress=True, local_cutoff=None):
        """Construct a handle for the local file 'filename'. This will
           create the initial version of the file that can be uploaded
           to the storage service. If the file is less than
           'local_cutoff' bytes then the file will be held directly
           in this handle. By default local_cutoff is 1 MB
        """
        self._local_filename = None
        self._local_filedata = None
        self._compression = None
        self._compressed_filename = None
        self._drive_uid = drive_uid
        self._aclrules = None

        if filename is not None:
            if local_cutoff is None:
                local_cutoff = 1048576
            else:
                local_cutoff = int(local_cutoff)

            if aclrules is None:
                # will be automatically set to 'inherit' on the service
                self._aclrules = None
            else:
                from Acquire.Identity import ACLRules as _ACLRules
                self._aclrules = _ACLRules.create(rule=aclrules)

            from Acquire.Access import get_filesize_and_checksum \
                as _get_filesize_and_checksum
            import os as _os

            (filesize, cksum) = _get_filesize_and_checksum(filename=filename)

            if compress and _should_compress(filename=filename,
                                             filesize=filesize):
                import bz2 as _bz2
                if filesize < local_cutoff:
                    # this is not big, so better to compress in memory
                    from Acquire.Access import get_size_and_checksum \
                        as _get_size_and_checksum
                    data = open(filename, "rb").read()
                    data = _bz2.compress(data)
                    (filesize, cksum) = _get_size_and_checksum(data=data)
                    self._local_filedata = data
                    self._compression = "bz2"
                else:
                    # this is a bigger file, so compress on disk
                    try:
                        self._compressed_filename = _bz2compress(
                                                        inputfile=filename)
                    except:
                        pass

                    if self._compressed_filename is not None:
                        self._compression = "bz2"
                        (filesize, cksum) = _get_filesize_and_checksum(
                                            filename=self._compressed_filename)
            elif filesize < local_cutoff:
                # this is small enough to hold in memory
                self._local_filedata = open(filename, "rb").read()

            if self._compressed_filename is None:
                self._local_filename = filename

            self._filesize = filesize
            self._checksum = cksum

            if remote_filename is None:
                self._filename = _os.path.split(filename)[1]
            else:
                from Acquire.ObjectStore import string_to_filepath \
                    as _string_to_filepath
                self._filename = _string_to_filepath(remote_filename)
        else:
            self._filename = None

    def __del__(self):
        """Ensure we delete the temporary file before being destroyed"""
        if self._compressed_filename is not None:
            import os as _os
            _os.unlink(self._compressed_filename)
            self._compressed_filename = None

    def __str__(self):
        """Return a string representation of the file"""
        if self.is_null():
            return "FileHandle::null"

        return "FileHandle(filename='%s')" % self.filename()

    def is_null(self):
        """Return whether or not this this null

        Returns:
            bool: True if handle null, else False

        """
        return self._filename is None

    def is_compressed(self):
        """Return whether or not the file is compressed on transport

           Returns:
                bool: True if compressed, else False
        """
        return self._compression is not None

    def compression_type(self):
        """Return a string describing the compression scheme used by the
           filehandle when transporting the file, or None if the data
           is not compressed

           Returns:
                str: Compression type
        """
        return self._compression

    def is_localdata(self):
        """Return whether or not this file is so small that the data
           is held in memory

           Returns:
                bool: True if file held in memory, else False
        """
        return self._local_filedata is not None

    def local_filedata(self, decompress=False):
        """Return the filedata for this file, assuming it is sufficiently
           small to be read in this way. Returns 'None' if not...

           If 'decompress' is true, then decompress the data
           (if it is compressed) before returning

           Args:
                decompress (bool, default=False): If True decompress
                data locally
           Returns:
                None or str: If decompress=True, return None, else return
                filedata
        """
        if decompress and self.is_compressed():
            if self._local_filedata is not None:
                import bz2 as _bz2
                return _bz2.decompress(self._local_filedata)
            else:
                return None
        else:
            return self._local_filedata

    def local_filename(self):
        """Return the local filename for this file

           Returns:
                str: Local filename for file
        """
        if self.is_localdata():
            return None
        elif self.is_compressed():
            return self._compressed_filename
        else:
            return self._local_filename

    def drive_uid(self):
        """Return the UID of the drive on which this file is located

           Returns:
                str: UID for drive
        """
        return self._drive_uid

    def aclrules(self):
        """Return the ACL rules for this file

           Returns:
                str: ACL rules for file
        """
        return self._aclrules

    def filename(self):
        """Return the remote (object store) filename for this file

           Returns:
                str: Filename
        """
        return self._filename

    def filesize(self):
        """Return the size (in bytes) of this file

           Returns:
                int: Size of file in bytes
        """
        if self.is_null():
            return 0
        else:
            return self._filesize

    def checksum(self):
        """Return the checksum of the contents of this file

           Returns:
                str: MD5 checksum for file
        """
        if self.is_null():
            return None
        else:
            return self._checksum

    def fingerprint(self):
        """Return a fingerprint for this file

           Returns:
                str: Fingerprint for file consisting of
                filename, file size and MD5 checksum
        """
        return "%s:%s:%s" % (self.filename(),
                             self.filesize(), self.checksum())

    def to_data(self):
        """Return a json-serialisable dictionary for this object. Note
           that this does not contain any information about the local
           file itself - just the name it should be called on the
           object store and the size, checksum and acl. If the file
           (or compressed file) is sufficiently small then this
           will also contain the packed version of that file data

           Returns:
                dict: JSON serialisable dictionary of object
        """
        data = {}

        if not self.is_null():
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            data["filename"] = self.filename()
            data["filesize"] = self.filesize()
            data["checksum"] = self.checksum()

            if self._aclrules is not None:
                data["aclrules"] = self._aclrules.to_data()

            data["drive_uid"] = self.drive_uid()

            if self._local_filedata is not None:
                from Acquire.ObjectStore import bytes_to_string \
                    as _bytes_to_string
                data["filedata"] = _bytes_to_string(self._local_filedata)

            if self._compression is not None:
                data["compression"] = self._compression

        return data

    @staticmethod
    def from_data(data):
        """Return an object created from the passed json-deserialised
           dictionary. Note that this does not contain any information
           about the local file itself - just the name it should be
           called on the object store and the size, checksum and acl.
           If the file (or compressed file) is sufficiently small then this
           will also contain the packed version of that file data

           Args:
                data (dict): JSON-deserialised dictionary
           Returns:
                FileHandle: FileHandle object created from dictionary
        """
        f = FileHandle()

        if data is not None and len(data) > 0:
            from Acquire.Storage import ACLRule as _ACLRule
            f._filename = data["filename"]
            f._filesize = int(data["filesize"])
            f._checksum = data["checksum"]
            f._drive_uid = data["drive_uid"]

            if "compression" in data:
                f._compression = data["compression"]

            if "aclrules" in data:
                from Acquire.Storage import ACLRules as _ACLRules
                f._aclrules = _ACLRules.from_data(data["aclrules"])

            if "filedata" in data:
                from Acquire.ObjectStore import string_to_bytes \
                    as _string_to_bytes

                f._local_filedata = _string_to_bytes(data["filedata"])

        return f




from cachetools import cached as _cached
from cachetools import LRUCache as _LRUCache

_cache_local_serviceinfo = _LRUCache(maxsize=5)
_cache_local_serviceinfos = _LRUCache(maxsize=5)

__all__ = ["get_trusted_service", "get_trusted_services",
           "clear_services_cache"]


def clear_services_cache():
    """Clear the caches of loaded services"""
    _cache_local_serviceinfo.clear()
    _cache_local_serviceinfos.clear()


# Cached as the list of all trusted remote service information
# will not change too often
@_cached(_cache_local_serviceinfos)
def get_trusted_services():
    """Return a dictionary of all trusted services indexed by
       their type
    """
    from Acquire.Service import is_running_service as _is_running_service

    if _is_running_service():
        from Acquire.Service import get_this_service as _get_this_service
        from Acquire.Service import Service as _Service

        from Acquire.Service import get_service_account_bucket as \
            _get_service_account_bucket
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import url_to_encoded as \
            _url_to_encoded

        # we already trust ourselves
        service = _get_this_service()

        trusted_services = {}
        trusted_services[service.service_type()] = [service]

        bucket = _get_service_account_bucket()

        uidkey = "_trusted/uid/"
        datas = _ObjectStore.get_all_objects(bucket, uidkey)

        for data in datas:
            remote_service = _Service.from_data(data)

            if remote_service.should_refresh_keys():
                # need to update the keys in our copy of the service
                remote_service.refresh_keys()
                key = "%s/%s" % (uidkey, remote_service.uid())
                _ObjectStore.set_object_from_json(bucket, key,
                                                  remote_service.to_data())

            if remote_service.service_type() in datas:
                datas[remote_service.service_type()].append(remote_service)
            else:
                datas[remote_service.service_type()] = [remote_service]

        return datas
    else:
        # this is running on the client
        from Acquire.Client import Wallet as _Wallet
        wallet = _Wallet()
        return wallet.get_services()


# Cached as the remote service information will not change too often
@_cached(_cache_local_serviceinfo)
def get_trusted_service(service_url=None, service_uid=None,
                        service_type=None, autofetch=True):
    """Return the trusted service info for the service with specified
       service_url or service_uid"""
    if service_url is not None:
        from Acquire.Service import Service as _Service
        service_url = _Service.get_canonical_url(service_url,
                                                 service_type=service_type)

    from Acquire.Service import is_running_service as _is_running_service

    if _is_running_service():
        from Acquire.Service import get_this_service as _get_this_service
        from Acquire.Service import Service as _Service
        from Acquire.Service import get_service_account_bucket as \
            _get_service_account_bucket
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import url_to_encoded as \
            _url_to_encoded

        service = _get_this_service()

        if service_url is not None and service.canonical_url() == service_url:
            # we trust ourselves :-)
            return service

        if service_uid is not None and service.uid() == service_uid:
            # we trust ourselves :-)
            return service

        bucket = _get_service_account_bucket()
        uidkey = None
        data = None

        if service_uid is not None:
            uidkey = "_trusted/uid/%s" % service_uid
            try:
                data = _ObjectStore.get_object_from_json(bucket, uidkey)
            except:
                pass
        elif service_url is not None:
            urlkey = "_trusted/url/%s" % _url_to_encoded(service_url)
            try:
                uidkey = _ObjectStore.get_string_object(bucket, urlkey)
                if uidkey is not None:
                    data = _ObjectStore.get_object_from_json(bucket, uidkey)
            except:
                pass

        if data is not None:
            remote_service = _Service.from_data(data)

            if remote_service.should_refresh_keys():
                # need to update the keys in our copy of the service
                remote_service.refresh_keys()

                if uidkey is not None:
                    _ObjectStore.set_object_from_json(
                                                    bucket, uidkey,
                                                    remote_service.to_data())

            return remote_service

        if not autofetch:
            from Acquire.Service import ServiceAccountError
            if service_uid is not None:
                raise ServiceAccountError(
                    "We do not trust the service with UID '%s'" %
                    service_uid)
            else:
                raise ServiceAccountError(
                    "We do not trust the service at URL '%s'" %
                    service_url)

        # we can try to fetch this data - we will ask our own
        # registry
        from Acquire.Registry import get_trusted_registry_service \
            as _get_trusted_registry_service
        registry = _get_trusted_registry_service(service_uid=service.uid())
        service = registry.get_service(service_uid=service_uid,
                                       service_url=service_url)

        from Acquire.Service import trust_service as _trust_service
        _trust_service(service)
        return service
    else:
        # this is running on the client
        from Acquire.Client import Wallet as _Wallet
        wallet = _Wallet()
        service = wallet.get_service(service_uid=service_uid,
                                     service_url=service_url,
                                     service_type=service_type,
                                     autofetch=autofetch)
        return service




import json as _json
import base64 as _base64
import datetime as _datetime
import uuid as _uuid
import os as _os
import sys as _sys

if _sys.version_info.major < 3:
    raise ImportError("Acquire requires Python 3.6 minimum")

if _sys.version_info.minor < 6:
    raise ImportError("Acquire requires Python 3.6 minimum")

if _sys.version_info.major == 3 and _sys.version_info.minor == 6:
    try:
        from backports.datetime_fromisoformat import MonkeyPatch \
            as _MonkeyPatch
        _MonkeyPatch.patch_fromisoformat()
    except:
        raise ImportError(
            "backports-datetime-fromisoformat must be installed "
            "on Python < 3.7. Please run 'pip install "
            "backports-datetime-fromisoformat' or update to a newer "
            "version of Python")

__all__ = ["bytes_to_string", "string_to_bytes",
           "string_to_encoded", "encoded_to_string",
           "url_to_encoded", "encoded_to_url",
           "decimal_to_string", "string_to_decimal",
           "datetime_to_string", "string_to_datetime",
           "date_to_string", "string_to_date",
           "time_to_string", "string_to_time",
           "get_datetime_now", "datetime_to_datetime",
           "string_to_safestring", "safestring_to_string",
           "string_to_list", "list_to_string",
           "string_to_dict", "dict_to_string",
           "string_to_filepath", "string_to_filepath_parts",
           "get_datetime_future",
           "get_datetime_now_to_string",
           "date_and_time_to_datetime",
           "date_and_hour_to_datetime",
           "create_uuid", "create_uid", "validate_is_uid"]


def create_uuid(short_uid=False, include_date=None, separator="/"):
    """Return a newly created random uuid. This is highly likely
       to be globally unique. If 'short_uid' is True, then a shorter,
       potentially less unique UID will be generated. If
       'include_date' is passed, then the passed date will
       be encoded into the UID

       Returns:
            str: Random UUID
    """
    uid = str(_uuid.uuid4())

    if short_uid:
        uid = uid[:8]

    if include_date is not None and include_date is not False:
        if include_date is True:
            include_date = get_datetime_now()
        else:
            include_date = datetime_to_datetime(include_date)

        uid = "%s%s%s" % (include_date.replace(tzinfo=None).isoformat(),
                          separator, uid)

    return uid


def create_uid(short_uid=False, include_date=None, separator="/"):
    """Synonym for create_uuid"""
    return create_uuid(short_uid=short_uid, include_date=include_date,
                       separator=separator)


def validate_is_uid(uid):
    """Validate that the passed 'uid' is actually a UID. This checks
       that the string is not something weird that is trying to
       break the system
    """
    if uid is None:
        raise TypeError("'None' is not a valid UID!")

    uid = str(uid)

    len_uid = len(uid)

    import re as _re
    from Acquire.ObjectStore import string_to_datetime \
        as _string_to_datetime

    if len_uid == 8:
        # this is a short UID
        if _re.match(r'[a-f0-9]{8}', uid):
            return
    elif len_uid == 35:
        # this is a short UID with a datetime
        parts = uid.split("/")
        try:
            dt = _string_to_datetime(parts[0])
            if _re.match(r'[a-f0-9]{8}', parts[1]):
                return
        except Exception as e:
            print(e)
            pass
    elif len_uid == 36:
        # this is a long UID
        if _re.match(r'[a-f0-9]{8}\-[a-f0-9]{4}\-[a-f0-9]{4}\-[a-f0-9]{4}\-[a-f0-9]{12}', uid):
            return
    elif len_uid == 63:
        # this is a long UID with a datetime
        parts = uid.split("/")
        try:
            dt = _string_to_datetime(parts[0])
            if _re.match(r'[a-f0-9]{8}\-[a-f0-9]{4}\-[a-f0-9]{4}\-[a-f0-9]{4}\-[a-f0-9]{12}', parts[1]):
                return
        except:
            pass

    raise TypeError("'%s' is not a valid UID!" % uid)


def string_to_encoded(s):
    """Return the passed unicode string encoded to a safely
       encoded base64 utf-8 string

       Args:
            s (str): String to encode
       Returns:
            str: Unicode string s encoded to UTF-8

    """
    return bytes_to_string(s.encode("utf-8"))


def encoded_to_string(b):
    """Return the passed encoded base64 utf-8 string converted
       back into a unicode string

       Args:
            b (bytes): base64 utf-8 byte string
       Returns:
            str: UTF-8 string converted from byte string

    """
    return string_to_bytes(b).decode("utf-8")


def url_to_encoded(url):
    """Return an encoding of the passed url that is safe to use
       as a name, filename or key in an object store

       Args:
            url (str): URL to encode
       Returns:
            bytes: base64 encoded bytes object

    """
    return _base64.b64encode(url.encode("utf-8")).decode("utf-8")


def encoded_to_url(e):
    """Decode the passed encoded data back to the URL. This will only
       produce valid output for inputs created using url_to_encoded

       Args:
            e (bytes): base64 encoded byte object
       Returns:
            bytes: bytes object as URL
    """
    return _base64.b64decode(e.encode("utf-8")).decode("utf-8")


def bytes_to_string(b):
    """Return the passed binary bytes safely encoded to
       a base64 utf-8 string

       Args:
            b (bytes): binary bytes to encode
       Returns:
            str: UTF-8 encoded string object

    """
    if b is None:
        return None
    else:
        return _base64.b64encode(b).decode("utf-8")


def string_to_bytes(s):
    """Return the passed base64 utf-8 encoded binary data
       back converted from a string back to bytes. Note that
       this can only convert strings that were encoded using
       bytes_to_string - you cannot use this to convert
       arbitrary strings to bytes

       Args:
            s (str): base64 byte object to decode
       Returns:
            bytes: bytes object

    """
    if s is None:
        return None
    else:
        return _base64.b64decode(s.encode("utf-8"))


def decimal_to_string(d):
    """Return the passed decimal number encoded as a string that
       can be safely serialised via JSON

       Args:
            d (Decimal): Decimal to convert to string
       Returns:
            str: String version of Decimal
    """
    return str(d)


def string_to_decimal(s, default=0):
    """Return the decimal that had been encoded via 'decimal_to_string'.
       This string must have been created via 'decimal_to_string'

       Args:
            s (str): String to convert to Decimal
       Returns:
            Decimal: Decimal version of string
    """
    from Acquire.Accounting import create_decimal as _create_decimal
    return _create_decimal(s, default=default)


def datetime_to_string(d):
    """Return the passed datetime encoded to a string. This will be a
       standard iso-formatted time in the UTC timezone (converting
       to UTC if the passed datetime is for another timezone)

       Args:
            d (datetime): Datetime to convert to string
       Returns:
            str: Datetime as a string

    """
    if d.tzinfo is None:
        d = d.replace(tzinfo=_datetime.timezone.utc)
    else:
        d = d.astimezone(_datetime.timezone.utc)

    # the datetime is in UTC, so write out the string without
    # the unnecessary +00:00
    return d.replace(tzinfo=None).isoformat()


def datetime_to_datetime(d):
    """Return the passed datetime as a datetime that is clean
       and usable by Acquire. This will move the datetime to UTC,
       adding the timezone if this is missing

       Args:
            d (datetime): datetime to convert to UTC
       Returns:
            datetime: UTC datetime useable by Acquire

    """
    if not isinstance(d, _datetime.datetime):
        raise TypeError(
            "The passed object '%s' is not a valid datetime" % str(d))

    if d.tzinfo is None:
        return d.replace(tzinfo=_datetime.timezone.utc)
    else:
        return d.astimezone(_datetime.timezone.utc)


def date_and_time_to_datetime(date, time=_datetime.time(0)):
    """Return the passed date and time as a UTC datetime. By
       default the time is midnight (first second of the day)

       Args:
            date (datetime): Date (may be any timezone)
            time (default=_datetime.time(0)): Time of day, default midnight
       Returns:
            datetime: UTC datetime
    """
    return datetime_to_datetime(_datetime.datetime.combine(date, time))


def date_and_hour_to_datetime(date, hour):
    """Return the passed date and hour as a UTC datetime. By
       default the time is hour:00:00 (first second of the hour)
    """
    return datetime_to_datetime(
                _datetime.datetime.combine(date,
                                           _datetime.time(hour=hour)))


def get_datetime_now():
    """Return the current time in the UTC timezone. This creates an
       object that will be properly stored using datetime_to_string
       and string_to_datetime

       Returns:
            datetime: Current datetime
    """
    return datetime_to_datetime(_datetime.datetime.now(
                                _datetime.timezone.utc))


def get_datetime_now_to_string():
    """Convenience function that returns the result of get_datetime_now
       as a string converted via datetime_to_string

       Returns:
            str: Current datetime as string
    """
    return datetime_to_string(get_datetime_now())


def get_datetime_future(weeks=0, days=0, hours=0, minutes=0, seconds=0,
                        timedelta=None):
    """Return the datetime that is the supplied time in the future.
       This will raise an exception if the time is not in the future!

       Args:
            weeks (int, default=0): Number of weeks in future
            days (int, default=0): Number of days in future
            hours (int, default=0): Number of hours in future
            minutes (int, default=0): Number of minutes in future
            seconds (int, default=0): Number of seconds in future
            timedelta (datetime.timedelta, default=0): Timedelta from now
    """
    delta = _datetime.timedelta(weeks=weeks, days=days, hours=hours,
                                minutes=minutes, seconds=seconds)

    if timedelta is not None:
        if not isinstance(timedelta, _datetime.timedelta):
            raise TypeError("The delta must be a datetime.timedelta object")

        delta += timedelta

    if delta.total_seconds() < 5:
        raise ValueError(
            "The requested delta (%s) is not sufficiently far enough "
            "into the future!" % str(delta))

    return get_datetime_now() + delta


def string_to_datetime(s):
    """Return the datetime that had been encoded to the passed string
       via datetime_to_string. This string must have been created
       via 'datetime_to_string'

       Args:
            s (str): String to convert
       Returns:
            datetime: Datetime version of string
    """
    if isinstance(s, _datetime.datetime):
        return s
    else:
        d = _datetime.datetime.fromisoformat(s)
        return datetime_to_datetime(d)


def date_to_string(d):
    """Return the date that has been encoded to a string. This will
       write the date as a standard iso-formatted date. IF a datetime
       is passed then this will be in the
       UTC timezone (converting to UTC if the passed datetime
       is for another timezone)

       Args:
            d (datetime): Datetie to convert
       Returns:
            str: Datetime in ISO format
    """
    if isinstance(d, _datetime.datetime):
        return datetime_to_datetime(d).date().isoformat()
    else:
        return d.isoformat()


def string_to_date(s):
    """Return a date from the string that has been encoded using
       'date_to_string'. This is only guaranteed to work for strings
       that were created using that function

       Args:
            s (str): String from date_to_string function
       Returns:
            datetime: Datetime created from string
    """
    d = _datetime.date.fromisoformat(s)
    return d


def time_to_string(t):
    """Return the time that has been encoded to a string. This will
       write the time as a standard iso-formatted time. If a datetime
       is passed then this will be in the
       UTC timezone (converting to UTC if the passed datetime
       is for another timezone)

       Args:
            t (time): Time to convert, can be datetime
       Returns:
            str: String of passed time converted to UTC ISO
            format
    """
    if isinstance(t, _datetime.datetime):
        t = datetime_to_datetime(t)

        # guaranteed to be in the utc timezone, so write the
        # time without the unnecessary +00:00
        return t.replace(tzinfo=None).time().isoformat()
    else:
        if t.tzinfo is None:
            # assume UTC
            t = t.replace(tzinfo=_datetime.timezone.utc)
        elif t.tzinfo != _datetime.timezone.utc:
            from Acquire.ObjectStore import EncodingError
            raise EncodingError(
                "Cannot encode a time to a string as this time is "
                "not in the UTC timezone. Please convert to UTC "
                "before encoding this time to a string '%s'" % t.isoformat())

        # as the time is in UTC, we don't need the unnecessary +00:00
        return t.replace(tzinfo=None).isoformat()


def string_to_time(s):
    """Return a time from the string that was encoded by 'time_to_string'.
       This will only be guaranteed to produce valid output for strings
       produced using that function

       Args:
            s (str): String to convert to time
       Returns:
            datetime.time: Time object created from string
    """
    t = _datetime.time.fromisoformat(s)

    if t.tzinfo is None:
        # assume this is a UTC time
        t = t.replace(tzinfo=_datetime.timezone.utc)
    else:
        t = t.astimezone(_datetime.timezone.utc)

    return t


def string_to_safestring(s):
    """Return a safe (base64) encoded version of 's'. This string
       has no special characters or spaces, thereby making it safe
       for use, e.g. as a filename or to save in a database

       Args:
            s (str): String to encode to base
       Returns:
            str: base64 encoded string
    """
    return _base64.b64encode(s.encode("utf-8")).decode("utf-8")


def safestring_to_string(s):
    """Return the original string encoded by string_to_safestring"""
    return _base64.b64decode(s.encode("utf-8")).decode("utf-8")


def list_to_string(l):
    """Return the passed list of items converted to a json string.
       All items should have the same type

       Args:
            l (list): List to convert to JSON
       Returns:
            str: JSON string
    """
    j = []
    for item in l:
        j.append(item.to_data())

    return _json.dumps(j)


def string_to_list(s, C):
    """Convert the string encoded using list_to_string back to a list
        of objects of type C. Note that all objects must have the
        same type

        Args:
            s (str): String to convert
            C (type): Type to create
        Returns:
            list: List of C objects
    """
    items = []

    for val in _json.loads(s):
        items.append(C.from_data(val))

    return items


def dict_to_string(d):
    """Return the passed dict of items converted to a json string.
       All items should have the same type

       Args:
            d (dict): Dictionary to convert
       Returns:
            str: JSON version of dict
    """
    j = {}
    for key, value in d.items():
        if value is None:
            j[key] = None
        else:
            j[key] = value.to_data()

    return _json.dumps(j)


def string_to_dict(s, C):
    """Convert the string encoded using dict_to_string back to a dict
        of objects of type C. Note that all objects must have the
        same type

        Args:
            s (str): String to convert
            C (type): Type to create
        Returns:
            dict: Dictionary created from string
    """
    items = {}

    for key, value in _json.loads(s).items():
        if value is None:
            items[key] = None
        else:
            items[key] = C.from_data(value)

    return items


def string_to_filepath(path):
    """This function cleans the passed path so that doesn't contain
       redundant slashes or '..' etc., so that all backslashes are forwards
       slashes, and that the trailing slash is removed

       Args:
            path (str): Path to clean
       Returns:
            str: Cleaned path
    """
    if path is None:
        return ""

    # change any windows path separators into unix path separators
    path = path.replace("\\", "/")

    # now normalise the path to remove messiness
    path = _os.path.normpath(path)

    # remove all ".." and "." from this path
    if path.find(".") != -1:
        parts = path.split("/")
        for i, part in enumerate(parts):
            if part == ".":
                parts[i] = None
            elif part == "..":
                if i == 0:
                    raise ValueError("You cannot start a path with '..'")
                part[i-1] = None
                part[i] = None

        path = _os.path.normpath("/".join(parts))

    return path


def string_to_filepath_parts(path):
    """Break the passed path into a list of the individual parts,
       e.g. /home/user/test/../something/./new.txt will return

       ['home', 'user', 'something', 'new.txt']

       Args:
            path (str): Path to split
       Returns:
            list: List of sections of path
    """
    from os.path import split as _split

    path = string_to_filepath(path)

    dirs = []

    (root, part) = _split(path)

    dirs.append(part)

    while len(root) > 0:
        (root, part) = _split(root)

        if len(part) > 0:
            dirs.insert(0, part)
        else:
            break

    return dirs




__all__ = ["WorkSheet"]


class WorkSheet:
    """This class holds a complete record of the work that the access
       service has been asked to perform.
    """
    def __init__(self, request=None, authorisation=None):
        if request is not None:
            from Acquire.Identity import Authorisation as _Authorisation
            if not isinstance(authorisation, _Authorisation):
                raise TypeError("You can only authorise work with "
                                "a valid Authorisation object")

            from Acquire.Access import RunRequest as _RunRequest
            if not isinstance(request, _RunRequest):
                raise TypeError(
                    "You must pass in a valid RunRequest to request a "
                    "calculation is run. The passed request is the wrong "
                    "type: %s" % str(request))

            authorisation.verify(request.fingerprint())
            from Acquire.ObjectStore import create_uid as _create_uid
            self._request = request
            self._authorisation = authorisation
            self._uid = _create_uid(include_date=True, short_uid=True,
                                    separator="-")
            self._status = "awaiting"
        else:
            self._uid = None
            self._status = None

        self._credit_notes = None

    def is_null(self):
        """Return whether or not this WorkSheet is null"""
        return self._uid is None

    def uid(self):
        """Return the UID of this WorkSheet"""
        return self._uid

    def status(self):
        """Return the status of the work"""
        if self.is_null():
            return None
        else:
            return self._status

    def storage_service(self):
        """Return the storage service that will be used to store
           the output data associated with this work

           TODO - will eventually have to choose which storage service
                  to use. Currently returning the storage service that
                  is at the same root URL as this access service
        """
        from Acquire.Service import get_this_service as _get_this_service
        from Acquire.Service import get_trusted_service as _get_trusted_service
        service = _get_this_service()
        storage_url = service.canonical_url().replace("access", "storage")
        return _get_trusted_service(service_url=storage_url)

    def compute_service(self):
        """Return the compute service that will be used to actually
           perform the calculation associated with this work

           TODO - will eventually have to choose which compute service
                  to use. Currently returning the compute service that
                  is at the same root URL as this access service
        """
        from Acquire.Service import get_this_service as _get_this_service
        from Acquire.Service import get_trusted_service as _get_trusted_service
        service = _get_this_service()
        compute_url = service.canonical_url().replace("access", "compute")
        return _get_trusted_service(service_url=compute_url)

    def total_cost(self):
        """Return the total maximum quoted cost for this work. The
           total cost to run the work must not exceed this

           TODO - will need to actually work out and return the
                  real cost - returning dummy values for the moment
        """
        if self.is_null():
            return 0
        else:
            return 10

    def request(self):
        """Return the original work request"""
        if self.is_null():
            return None
        else:
            return self._request

    def authorisation(self):
        """Return the original authorisation for this request"""
        if self.is_null():
            return None
        else:
            return self._authorisation

    def user_guid(self):
        """Return the GUID of the user who requested this work"""
        if self.is_null():
            return None
        else:
            return self._authorisation.user_guid()

    def execute(self, cheque):
        """Execute (start) this work, using the passed cheque for
           payment. Note that you can't perform the same work twice
        """
        if self.is_null():
            from Acquire.Accounting import PaymentError
            raise PaymentError("You cannot try to execute null work!")

        from Acquire.Client import Cheque as _Cheque
        if not isinstance(cheque, _Cheque):
            raise TypeError("You must pass a valid Cheque as payment "
                            "for the work")

        if self._credit_notes is not None:
            raise PermissionError("You cannot start a piece of work twice!")

        from Acquire.Service import get_this_service as _get_this_service

        access_service = _get_this_service(need_private_access=True)
        compute_service = self.compute_service()
        storage_service = self.storage_service()
        accounting_service = cheque.accounting_service()

        access_user = access_service.login_service_user()

        account_uid = access_service.service_user_account_uid(
                                    accounting_service=accounting_service)

        from Acquire.Client import Account as _Account
        access_account = _Account(user=access_user, account_uid=account_uid,
                                  accounting_service=accounting_service)

        # TODO - validate that the cost of the work on the compute
        #        and storage services is covered by the passed cheque

        try:
            credit_notes = cheque.cash(spend=self.total_cost(),
                                       resource=self.request().fingerprint())
        except Exception as e:
            from Acquire.Service import exception_to_string
            from Acquire.Accounting import PaymentError
            raise PaymentError(
                "Problem cashing the cheque used to pay for the calculation: "
                "\n\nCAUSE: %s" % exception_to_string(e))

        if credit_notes is None or len(credit_notes) == 0:
            from Acquire.Accounting import PaymentError
            raise PaymentError("Cannot be paid!")

        # make sure that we have been paid!
        for credit_note in credit_notes:
            if credit_note.credit_account_uid() != access_account.uid():
                raise PaymentError("The wrong account has been paid!?!")

        self._status = "awaiting (paid)"
        self._credit_notes = credit_notes

        # work out when this job MUST have finished. If the job
        # has not completed before this time then it will be killed
        from Acquire.ObjectStore import get_datetime_future \
            as _get_datetime_future

        endtime = _get_datetime_future(days=2)  # this should be calculated

        # save the WorkSheet to the object store so we don't lose the
        # value in the credit notes
        self.save()

        compute_cheque = _Cheque.write(
                                account=access_account,
                                resource="work %s" % self.uid(),
                                max_spend=10.0,
                                recipient_url=compute_service.canonical_url(),
                                expiry_date=endtime)

        storage_cheque = _Cheque.write(
                                account=access_account,
                                resource="work %s" % self.uid(),
                                max_spend=10.0,
                                recipient_url=storage_service.canonical_url(),
                                expiry_date=endtime)

        self._compute_cheque = compute_cheque
        self._storage_cheque = storage_cheque

        # now create a Drive on the storage service that will hold
        # the output for this job
        from Acquire.Client import Drive as _Drive
        from Acquire.Client import StorageCreds as _StorageCreds
        from Acquire.Client import ACLRule as _ACLRule
        from Acquire.Client import ACLRules as _ACLRules
        from Acquire.Client import ACLUserRules as _ACLUserRules

        creds = _StorageCreds(user=access_user,
                              storage_service=storage_service)

        rule = _ACLUserRules.owner(user_guid=access_user.guid()).add(
                    user_guid=self.user_guid(), rule=_ACLRule.reader())

        aclrules = _ACLRules(rule=rule, default_rule=_ACLRule.denied())

        output_drive = _Drive(name="output_%s" % self.uid(),
                              creds=creds, aclrules=aclrules,
                              cheque=storage_cheque,
                              max_size="10MB",
                              autocreate=True)

        self._output_loc = output_drive.metadata().location()
        self._status = "awaiting (paid, have drive)"
        self.save()

        from Acquire.Client import PAR as _PAR
        par = _PAR(location=self._output_loc, user=access_user,
                   aclrule=_ACLRule.writer(),
                   expires_datetime=endtime)

        secret = compute_service.encrypt_data(par.secret())

        args = {"worksheet_uid": self.uid(),
                "request": self.request().to_data(),
                "par": par.to_data(),
                "secret": secret,
                "cheque": compute_cheque.to_data()}

        self._status = "submitting"
        self.save()

        response = compute_service.call_function(function="submit_job",
                                                 args=args)

        print(response)

        self._status = "submitted"
        self.save()

        # the service user will log out automatically on destruction, but
        # let us make sure!
        access_user.logout()

        # TODO - should collect something from the response that
        #        can be saved in the job sheet so that we know
        #        how submission is going...

    def output_location(self):
        """Return the location where the output will be saved"""
        if self.is_null():
            return None
        else:
            return self._output_loc

    def save(self):
        """Save this WorkSheet to the object store
            Returns:
                None
        """
        from Acquire.Service import assert_running_service \
            as _assert_running_service

        _assert_running_service()

        if self.is_null():
            return

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()

        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        key = "worksheet/%s" % self.uid()
        _ObjectStore.set_object_from_json(bucket, key, self.to_data())

    @staticmethod
    def load(uid):
        """Return the WorkSheet with specified uid loaded from the
           ObjectStore
        """
        from Acquire.Service import assert_running_service \
            as _assert_running_service

        _assert_running_service()

        if uid is None:
            return

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()

        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        key = "worksheet/%s" % str(uid)
        data = _ObjectStore.get_object_from_json(bucket, key)
        return WorkSheet.from_data(data)

    def to_data(self):
        """Get a JSON-serialisable dictionary of this object"""
        data = {}

        if self.is_null():
            return data

        from Acquire.ObjectStore import list_to_string as _list_to_string

        data["uid"] = self.uid()

        try:
            data["request"] = self.request().to_data()
        except:
            pass

        try:
            data["authorisation"] = self.authorisation().to_data()
        except:
            pass

        try:
            data["credit_notes"] = _list_to_string(self._credit_notes)
        except:
            pass

        data["status"] = self._status

        try:
            data["output_location"] = self._output_loc.to_string()
        except:
            pass

        try:
            data["compute_cheque"] = self._compute_cheque.to_data()
        except:
            pass

        try:
            data["storage_cheque"] = self._storage_cheque.to_data()
        except:
            pass

        return data

    @staticmethod
    def from_data(data):
        """Return a WorkSheet constructed from the passed JSON-deserialised
           dictionary
        """
        if data is None or len(data) == 0:
            return

        j = WorkSheet()

        from Acquire.Access import RunRequest as _RunRequest
        from Acquire.Client import Authorisation as _Authorisation
        from Acquire.Client import Location as _Location
        from Acquire.Client import Cheque as _Cheque
        from Acquire.Accounting import CreditNote as _CreditNote
        from Acquire.ObjectStore import string_to_list \
            as _string_to_list

        j._uid = str(data["uid"])

        if "request" in data:
            j._request = _RunRequest.from_data(data["request"])

        if "authorisation" in data:
            j._authorisation = _Authorisation.from_data(
                                            data["authorisation"])

        if "credit_notes" in data:
            j._credit_notes = _string_to_list(data["credit_notes"],
                                              _CreditNote)

        j._status = data["status"]

        if "output_location" in data:
            j._output_loc = _Location.from_string(data["output_location"])

        if "compute_cheque" in data:
            j._compute_cheque = _Cheque.from_data(data["compute_cheque"])

        if "storage_cheque" in data:
            j._storage_cheque = _Cheque.from_data(data["storage_cheque"])

        return j




__all__ = ["ComputeJob"]


class ComputeJob:
    """This class holds all information about a compute job. It is used
       as part of record keeping in the compute service, and to pass
       information between the compute service and the actual
       compute cluster
    """
    def __init__(self):
        self._uid = None

    def __str__(self):
        if self.is_null():
            return "ComputeJob::null"
        else:
            return "ComputeJob(uid=%s, request=%s, output=%s)" % (
                self._uid, self._request, self._par.location().to_string())

    def __repr__(self):
        return self.__str__()

    def is_null(self):
        return self._uid is None

    def uid(self):
        """Return the UID of this job"""
        return self._uid

    def request(self):
        """Return the request associated with this job"""
        if self.is_null():
            return None
        else:
            return self._request

    @staticmethod
    def submit(worksheet_uid, request, par, secret, cheque):
        """Submit a job which has;
           worksheet_uid -

        """
        from Acquire.Service import get_this_service as _get_this_service
        from Acquire.Access import RunRequest as _RunRequest
        from Acquire.Client import PAR as _PAR
        from Acquire.Client import Cheque as _Cheque
        from Acquire.ObjectStore import create_uid as _create_uid
        from Acquire.Compute import Cluster as _Cluster

        if not isinstance(request, _RunRequest):
            raise TypeError("The request must be type RunRequest")

        if not isinstance(par, _PAR):
            raise TypeError("The PAR must be type PAR")

        if not isinstance(cheque, _Cheque):
            raise TypeError("The cheque must be type Cheque")

        service = _get_this_service(need_private_access=True)
        cluster = _Cluster.get_cluster()

        job = ComputeJob()

        job._secret = cluster.encrypt_data(service.decrypt_data(secret))
        job._par = par
        job._request = request

        cost = 10  # TODO - calculate cost of job again from request

        try:
            credit_notes = cheque.cash(spend=cost,
                                       resource="work %s" % worksheet_uid)
        except Exception as e:
            from Acquire.Service import exception_to_string
            from Acquire.Accounting import PaymentError
            raise PaymentError(
                "Problem cashing the cheque used to pay for the calculation: "
                "\n\nCAUSE: %s" % exception_to_string(e))

        if credit_notes is None or len(credit_notes) == 0:
            from Acquire.Accounting import PaymentError
            raise PaymentError("Cannot be paid!")

        job._credit_notes = credit_notes

        job._uid = _create_uid(include_date=True, short_uid=True,
                               separator="/")

        job.save()

        # signal server running on cluster to fetch job and actually
        # submit to slurm
        cluster.submit_job(job._uid)

        return job

    def save(self):
        """Save this ComputeJob to the objectstore"""
        if self.is_null():
            return

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()
        key = "compute/job/%s" % self._uid

        _ObjectStore.set_object_from_json(bucket, key, self.to_data())

    @staticmethod
    def load(uid):
        """Load the ComputeJob associated with the specified uid"""
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()
        key = "compute/job/%s" % uid

        try:
            data = _ObjectStore.get_object_from_json(bucket, key)
        except:
            data = None

        if data is None:
            raise KeyError("There is no job with UID = %s" % uid)

        return ComputeJob.from_data(data)

    def to_data(self):
        """Serialise this job to a json-serialisable dictionary"""
        if self.is_null():
            return {}

        from Acquire.ObjectStore import list_to_string as _list_to_string

        data = {}
        data["uid"] = self._uid
        data["credit_notes"] = _list_to_string(self._credit_notes)
        data["par"] = self._par.to_data()
        data["secret"] = self._secret
        data["request"] = self._request.to_data()

        return data

    @staticmethod
    def from_data(data):
        """Return a ComputeJob constructed from a json-deserialised
           dictionary
        """
        if data is None or len(data) == 0:
            return ComputeJob()

        from Acquire.ObjectStore import string_to_list as _string_to_list
        from Acquire.Access import RunRequest as _RunRequest
        from Acquire.Client import PAR as _PAR
        from Acquire.Accounting import CreditNote as _CreditNote

        job = ComputeJob()

        job._uid = str(data["uid"])
        job._credit_notes = _string_to_list(data["credit_notes"], _CreditNote)
        job._par = _PAR.from_data(data["par"])
        job._secret = str(data["secret"])
        job._request = _RunRequest.from_data(data["request"])

        return job




from Acquire.Service import Service as _Service

__all__ = ["RegistryService"]


class RegistryService(_Service):
    """This is a specialisation of Service for Registry Services"""
    def __init__(self, other=None):
        if isinstance(other, _Service):
            from copy import copy as _copy
            self.__dict__ = _copy(other.__dict__)

            if not self.is_registry_service():
                from Acquire.Registry import RegistryServiceError
                raise RegistryServiceError(
                    "Cannot construct a RegistryService from "
                    "a service which is not a registry service!")
        else:
            _Service.__init__(self)

    def _call_local_function(self, function, args):
        """Internal function called to short-cut local 'remote'
           function calls
        """
        from registry.route import registry_functions as _registry_functions
        from admin.handler import create_handler as _create_handler
        handler = _create_handler(_registry_functions)
        return handler(function=function, args=args)

    def get_service(self, service_url=None, service_uid=None):
        """Ask the registry to return the service with specified
           service_url or service_uid
        """
        if service_url is None and service_uid is None:
            raise PermissionError(
                "You must specify one of service_url or service_uid")

        args = {"service_uid": service_uid,
                "service_url": service_url}

        response = self.call_function(function="get_service",
                                      args=args)

        from Acquire.Service import Service as _Service
        service = _Service.from_data(response["service_info"])

        return service




__all__ = ["get_registry_details", "update_registry_keys_and_certs"]

_testing_registry = {
    "canonical_url": "registry",
    "uid": "Z9-Z9",
    "public_key": None,
    "public_certificate": None}

_acquire_registry = {
    "canonical_url": "https://fn.acquire-aaai.com/t/registry",
    "uid": "a0-a0",
    "public_key": None,
    "public_certificate": None}

_registries = {_testing_registry["uid"]: _testing_registry,
               _acquire_registry["uid"]: _acquire_registry}


def update_registry_keys_and_certs(registry_uid, public_key,
                                   public_certificate):
    """This function is called to update the registry details stored
       globally with those in the newly-created registry-service. This
       function should only be called by registry services after
       construction
    """
    if registry_uid not in _registries:
        raise PermissionError(
            "Cannot update registry details as this is not one of the "
            "centrally approved registries!")

    from Acquire.Crypto import PublicKey as _PublicKey

    if type(public_key) is not _PublicKey:
        raise TypeError("The public key must be type PublicKey")

    if type(public_certificate) is not _PublicKey:
        raise TypeError("The public certificate must be type PublicKey")

    r = _registries[registry_uid]

    r["public_key"] = public_key.to_data()
    r["public_certificate"] = public_certificate.to_data()

    _registries[registry_uid] = r


def get_registry_details(registry_uid):
    """Return the details for the registry with specified UID.
       Note that this will only return details for the approved
       and centrally-registered registries. This returns
       a dictionary with key registry details.
    """
    try:
        registry = _registries[registry_uid]
    except:
        registry = _registries["a0-a0"]

    if registry["public_key"] is None:
        try:
            from importlib import import_module as _import_module
            _keys = _import_module("._keys_%s" % registry_uid,
                                   package="Acquire.Registry")
            registry["public_key"] = _keys.public_key
            registry["public_certificate"] = _keys.public_certificate
        except:
            pass

    import copy as _copy

    return _copy.copy(registry)
